import asyncio
import inspect
import datetime
import warnings
from typing import Union, List, Optional, Dict, Any, Callable, Type
from collections import defaultdict

import numpy as np
import pandas as pd
from onetick.py.otq import otq, pyomd, otli

from onetick import py as otp
from onetick.py import utils, configuration
from onetick.py.core.column_operations.base import _Operation
from onetick.py.types import datetime2timeval, datetime2expr
from onetick.py.core.source import _is_dict_required
from onetick.py.compatibility import (
    has_max_expected_ticks_per_symbol,
    has_password_param,
    has_query_encoding_parameter,
    _add_version_info_to_exception,
)
from onetick.py._stack_info import _add_stack_info_to_exception
from onetick.py.callback import LogCallback, ManualDataframeCallback


def run(query: Union[Callable, Dict, otp.Source, otp.MultiOutputSource,  # NOSONAR
                     otp.query, str, otq.EpBase, otq.GraphQuery,
                     otq.ChainQuery, otq.Chainlet, otq.SqlQuery, otp.SqlQuery],
        *,
        symbols: Union[List[Union[str, otq.Symbol]], otp.Source, str, None] = None,
        start: Union[datetime.datetime, otp.datetime, pyomd.timeval_t, None] = utils.adaptive,  # type: ignore
        end: Union[datetime.datetime, otp.datetime, pyomd.timeval_t, None] = utils.adaptive,  # type: ignore
        date: Union[datetime.date, otp.date, None] = None,
        start_time_expression: Optional[str] = None,
        end_time_expression: Optional[str] = None,
        timezone=utils.default,  # type: ignore
        context=utils.default,  # type: ignore
        username: Optional[str] = None,
        alternative_username: Optional[str] = None,
        password: Optional[str] = None,
        batch_size: Union[int, Type[utils.default], None] = utils.default,
        running: Optional[bool] = False,
        query_properties: Optional[pyomd.QueryProperties] = None,  # type: ignore
        concurrency: Union[int, Type[utils.default], None] = utils.default,
        apply_times_daily: Optional[int] = None,
        symbol_date: Union[datetime.datetime, int, str, None] = None,
        query_params: Optional[Dict[str, Any]] = None,
        time_as_nsec: bool = True,
        treat_byte_arrays_as_strings: bool = True,
        output_matrix_per_field: bool = False,
        output_structure: Optional[str] = None,
        return_utc_times: Optional[bool] = None,
        connection=None,
        callback=None,
        svg_path=None,
        use_connection_pool: bool = False,
        node_name: Union[str, List[str], None] = None,
        require_dict: bool = False,
        max_expected_ticks_per_symbol: Optional[int] = None,
        log_symbol: Union[bool, Type[utils.default]] = utils.default,
        encoding: Optional[str] = None,
        manual_dataframe_callback: bool = False,
        print_symbol_errors: bool = True):
    """
    Executes a query and returns its result.

    Parameters
    ----------
    query: :py:class:`onetick.py.Source`, otq.Ep, otq.GraphQuery, otq.ChainQuery, str, otq.Chainlet,\
            Callable, otq.SqlQuery, :py:class:`onetick.py.SqlQuery`
        Query to execute can be source, path of the query on a disk or onetick.query graph or event processor.
        For running OTQ files, it represents the path (including filename) to the OTQ file to run a single query within
        the file. If more than one query is present, then the query to be run must be specified
        (that is, ``'path_to_file/otq_file.otq::query_to_run'``).

        ``query`` can also be a function that has a symbol object as the first parameter.
        This object can be used to get symbol name and symbol parameters.
        Function must return a :py:class:`Source <onetick.py.Source>`.
    symbols: str, list of str, list of otq.Symbol, :py:class:`onetick.py.Source`, :pandas:`pandas.DataFrame`, optional
        Symbol(s) to run the query for passed as a string, a list of strings,
        a :pandas:`pandas.DataFrame` with the ``SYMBOL_NAME`` column,
        or as a "symbols" query which results include the ``SYMBOL_NAME`` column.
        The start/end times for the symbols query will taken from the params below.
        See :ref:`symbols <static/concepts/symbols:Symbols: bound and unbound>` for more details.
    start: :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`,\
            :py:class:`pyomd.timeval_t`, optional
        The start time of the query. Can be timezone-naive or timezone-aware. See also ``timezone`` argument.
        onetick.py uses :py:attr:`otp.config.default_start_time<onetick.py.configuration.Config.default_start_time>`
        as default value, if you don't want to specify start time, e.g. to use saved time of the query,
        then you should specify None value.
    end: :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`,\
          :py:class:`pyomd.timeval_t`, optional
        The end time of the query (note that it's non-inclusive).
        Can be timezone-naive or timezone-aware. See also ``timezone`` argument.
        onetick.py uses :py:attr:`otp.config.default_end_time<onetick.py.configuration.Config.default_end_time>`
        as default value, if you don't want to specify end time, e.g. to use saved time of the query,
        then you should specify None value.
    date: :py:class:`datetime.date`, :py:class:`otp.date <onetick.py.date>`, optional
        The date to run the query for. Can be set instead of ``start`` and ``end`` parameters.
        If set then the interval to run the query will be from 0:00 to 24:00 of the specified date.
    start_time_expression: str, :py:class:`~onetick.py.Operation`, optional
        Start time onetick expression of the query. If specified, it will take precedence over ``start``.
        Supported only if query is Source, Graph or Event Processor.
        Not supported for WebAPI mode.
    end_time_expression: str, :py:class:`~onetick.py.Operation`, optional
        End time onetick expression of the query. If specified, it will take precedence over ``end``.
        Supported only if query is Source, Graph or Event Processor.
        Not supported for WebAPI mode.
    timezone: str, optional
         The timezone of output timestamps.
         Also, when start and/or end arguments are timezone-naive, it will define their timezone.
         If parameter is omitted timestamps of ticks will be formatted
         with the default :py:attr:`otp.config.tz<onetick.py.configuration.Config.tz>`.
    context: str, optional
        Allows specification of different contexts from OneTick configuration to connect to.
        If not set then default :py:attr:`otp.config.context<onetick.py.configuration.Config.context>` is used.
        See :ref:`guide about switching contexts <switching contexts>` for examples.
    username
        The username to make the connection.
        By default the user which executed the process is used or the value specified in
        :py:attr:`otp.config.default_username<onetick.py.configuration.Config.default_username>`.
    alternative_username: str
        The username used for authentication.
        Needs to be set only when the tick server is configured to use password-based authentication.
        By default,
        :py:attr:`otp.config.default_auth_username<onetick.py.configuration.Config.default_auth_username>` is used.
        Not supported for WebAPI mode.
    password: str, optional
        The password used for authentication.
        Needs to be set only when the tick server is configured to use password-based authentication.
        Note: not supported and ignored on older OneTick versions.
        By default, :py:attr:`otp.config.default_password<onetick.py.configuration.Config.default_password>` is used.
    batch_size: int
        number of symbols to run in one batch.
        By default, the value from
        :py:attr:`otp.config.default_batch_size<onetick.py.configuration.Config.default_batch_size>` is used.
        Not supported for WebAPI mode.
    running: bool, optional
        Indicates whether a query is CEP or not. Default is `False`.
    query_properties: :py:class:`pyomd.QueryProperties` or dict, optional
       Query properties, such as ONE_TO_MANY_POLICY, ALLOW_GRAPH_REUSE, etc
    concurrency: int, optional
        The maximum number of CPU cores to use to process the query.
        By default, the value from
        :py:attr:`otp.config.default_concurrency<onetick.py.configuration.Config.default_concurrency>` is used.
    apply_times_daily: bool
        Runs the query for every day in the ``start``-``end`` time range,
        using the time components of ``start`` and ``end`` datetimes.

        Note that those daily intervals are executed separately, so you don't have access
        to the data from previous or next days (see example in the next section).
    symbol_date:
        The symbol date used to look up symbology mapping information in the reference database,
        expressed as datetime object or integer of YYYYMMDD format
    query_params: dict
        Parameters of the query.
    time_as_nsec: bool
        Outputs timestamps up to nanoseconds granularity
        (defaults to False: by default we output timestamps in microseconds granularity)
    treat_byte_arrays_as_strings: bool
        Outputs byte arrays as strings (defaults to True)
        Not supported for WebAPI mode.
    output_matrix_per_field: bool
        Changes output format to list of matrices per field.
        Not supported for WebAPI mode.
    output_structure: otp.Source.OutputStructure, optional

        Structure (type) of the result. Supported values are:
          - `df` (default) - the result is returned as :pandas:`pandas.DataFrame` object
            or dictionary of symbol names and :pandas:`pandas.DataFrame` objects
            in case of using multiple symbols or first stage query.
          - `map` - the result is returned as SymbolNumpyResultMap.
          - `list` - the result is returned as list.
          - `polars` - the result is returned as
            `polars.DataFrame <https://docs.pola.rs/api/python/stable/reference/dataframe/index.html>`_ object
            or dictionary of symbol names and dataframe objects
            (**Only supported in WebAPI mode**).
    return_utc_times: bool
        If True Return times in UTC timezone and in local timezone otherwise
        Not supported for WebAPI mode.
    connection: :py:class:`pyomd.Connection`
        The connection to be used for discovering nested .otq files
        Not supported for WebAPI mode.
    callback: :py:class:`onetick.py.CallbackBase`
         Class with callback methods.
         If set, the output of the query should be controlled with callbacks
         and this function returns nothing.
    svg_path: str, optional
        Not supported for WebAPI mode.
    use_connection_pool: bool
        Default is False. If set to True, the connection pool is used.
        Not supported for WebAPI mode.
    node_name: str, List[str], optional
        Name of the output node to select result from. If query graph has several output nodes, you can specify the name
        of the node to choose result from. If node_name was specified, query should be presented by path on the disk
        and output_structure should be `df`
    require_dict: bool
        If set to True, result will be forced to be a dictionary even if it's returned for a single symbol
    max_expected_ticks_per_symbol: int
        Expected maximum number of ticks per symbol (used for performance optimizations).
        By default, :py:attr:`otp.config.max_expected_ticks_per_symbol \
            <onetick.py.configuration.Config.max_expected_ticks_per_symbol>` is used.
        Not supported for WebAPI mode.
    log_symbol: bool
        Log currently executed symbol.
        Note that this only works with unbound symbols.
        Also in this case :py:func:`otp.run<onetick.py.run>` is executed in ``callback`` mode
        and no value is returned from the function, so it should be used only for debugging purposes.
        This logging will not work if some other value specified in parameter ``callback``.
        By default, :py:attr:`otp.config.log_symbol<onetick.py.configuration.Config.log_symbol>` is used.
    encoding: str, optional
        The encoding of string fields.
    manual_dataframe_callback: bool
        Create dataframe manually with ``callback`` mode.
        Only works if ``output_structure='df'`` is specified and parameter ``callback`` is not.
        May improve performance in some cases.
    print_symbol_errors_from_onetick: bool
        Applicable only when ``output_structure`` is set to *df*.
        Print symbol errors from OneTick as python warnings.

    Returns
    -------
    result, list, dict, :pandas:`pandas.DataFrame`, None
        result of the query

    Examples
    --------

    Running :py:class:`onetick.py.Source` and setting start and end times:

    >>> data = otp.Tick(A=1)
    >>> otp.run(data, start=otp.dt(2003, 12, 2), end=otp.dt(2003, 12, 4))
            Time  A
    0 2003-12-02  1

    Setting query interval with ``date`` parameter:

    >>> data = otp.Tick(A=1)
    >>> data['START'] = data['_START_TIME']
    >>> data['END'] = data['_END_TIME']
    >>> otp.run(data, date=otp.dt(2003, 12, 1))
            Time  A      START        END
    0 2003-12-01  1 2003-12-01 2003-12-02

    Running otq.Ep and passing query parameters:

    >>> ep = otq.TickGenerator(bucket_interval=0, fields='long A = $X').tick_type('TT')
    >>> otp.run(ep, symbols='LOCAL::', query_params={'X': 1})
            Time  A
    0 2003-12-04  1

    Running in callback mode:

    >>> class Callback(otp.CallbackBase):
    ...     def __init__(self):
    ...         self.result = None
    ...     def process_tick(self, tick, time):
    ...         self.result = tick
    >>> data = otp.Tick(A=1)
    >>> callback = Callback()
    >>> otp.run(data, callback=callback)
    >>> callback.result
    {'A': 1}

    Running with ``apply_times_daily``.
    Note that daily intervals are processed separately so, for example,
    we can't access column **COUNT** from previous day.

    >>> trd = otp.DataSource('US_COMP', symbols='AAPL', tick_type='TRD')  # doctest: +SKIP
    >>> trd = trd.agg({'COUNT': otp.agg.count()},
    ...               bucket_interval=12 * 3600, bucket_time='start')  # doctest: +SKIP
    >>> trd['PREV_COUNT'] = trd['COUNT'][-1]  # doctest: +SKIP
    >>> otp.run(trd, apply_times_daily=True,
    ...         start=otp.dt(2023, 4, 3), end=otp.dt(2023, 4, 5), timezone='EST5EDT')  # doctest: +SKIP
                     Time   COUNT  PREV_COUNT
    0 2023-04-03 00:00:00  328447           0
    1 2023-04-03 12:00:00  240244      328447
    2 2023-04-04 00:00:00  263293           0
    3 2023-04-04 12:00:00  193018      263293

    Using a function as a ``query``, accessing symbol name and parameters:

    >>> def query(symbol):
    ...     t = otp.Tick(X='x')
    ...     t['SYMBOL_NAME'] = symbol.name
    ...     t['SYMBOL_PARAM'] = symbol.PARAM
    ...     return t
    >>> symbols = otp.Ticks({'SYMBOL_NAME': ['A', 'B'], 'PARAM': [1, 2]})
    >>> result = otp.run(query, symbols=symbols)
    >>> result['A']
            Time  X SYMBOL_NAME  SYMBOL_PARAM
    0 2003-12-01  x           A             1
    >>> result['B']
            Time  X SYMBOL_NAME  SYMBOL_PARAM
    0 2003-12-01  x           B             2

    Debugging unbound symbols with ``log_symbol`` parameter:

    >>> data = otp.Tick(X=1)
    >>> symbols = otp.Ticks({'SYMBOL_NAME': ['A', 'B'], 'PARAM': [1, 2]})
    >>> otp.run(query, symbols=symbols, log_symbol=True)  # doctest: +ELLIPSIS
    Running query <onetick.py.sources.ticks.Tick object at ...>
    Processing symbol A
    Processing symbol B

    By default, some non-standard characters in data strings could be processed incorrectly:

    >>> data = ['AA測試AA']
    >>> source = otp.Ticks({'A': data})
    >>> otp.run(source)
            Time           A
    0 2003-12-01  AAæ¸¬è©¦AA

    To fix this you can pass `encoding` parameter to `otp.run`:

    .. testcode::
       :skipif: not has_query_encoding_parameter()

       data = ['AA測試AA']
       source = otp.Ticks({'A': data})
       df = otp.run(source, encoding="utf-8")
       print(df)

    .. testoutput::

               Time        A
       0 2003-12-01  AA測試AA

    Note that query ``start`` time is inclusive, but query ``end`` time is not,
    meaning that ticks with timestamps equal to the query end time will not be included:

    >>> data = otp.Tick(A=1, bucket_interval=24*60*60)
    >>> data['A'] = data['TIMESTAMP'].dt.day_of_month()
    >>> otp.run(data, start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 4))
            Time  A
    0 2003-12-01  1
    1 2003-12-02  2
    2 2003-12-03  3
    >>> otp.run(data, start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 2))
            Time  A
    0 2003-12-01  1

    If you want to include such ticks, you can add one nanosecond to the query end time:

    >>> otp.run(data, start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 2) + otp.Nano(1))
            Time  A
    0 2003-12-01  1
    1 2003-12-02  2
    """
    _ = otli.OneTickLib()

    query_schema = None
    if isinstance(query, otp.Source):
        query_schema = query.schema

    if timezone is utils.default:
        timezone = configuration.config.tz
    if context is utils.default or context is None:
        context = configuration.config.context
    if concurrency is utils.default:
        concurrency = configuration.default_query_concurrency()

    if batch_size is utils.default:
        batch_size = configuration.config.default_batch_size
    if query_properties is None:
        query_properties = pyomd.QueryProperties()

    if isinstance(query_properties, dict):
        qp_dict = query_properties
        query_properties = utils.query_properties_from_dict(qp_dict)
    else:
        qp_dict = utils.query_properties_to_dict(query_properties)

    if 'USE_FT' not in qp_dict:
        query_properties.set_property_value('USE_FT', otp.config.default_fault_tolerance)  # type: ignore[union-attr]

    if 'IGNORE_TICKS_IN_UNENTITLED_TIME_RANGE' not in qp_dict:
        query_properties.set_property_value('IGNORE_TICKS_IN_UNENTITLED_TIME_RANGE',  # type: ignore[union-attr]
                                            str(otp.config.ignore_ticks_in_unentitled_time_range).upper())

    if date is not None:
        for v in (start, end, start_time_expression, end_time_expression):
            if v is not None and v is not utils.adaptive:
                raise ValueError("Can't use 'date' parameter when other time interval parameters are specified")
        start = otp.date(date)
        end = start + otp.Day(1)

    has_source_start, has_source_end = False, False
    if isinstance(query, otp.Source):
        has_source_start, has_source_end = query.has_start_end_time()

    if (start is None or start is utils.adaptive) and otp.config.get('default_start_time') is None and \
            not has_source_start:
        warnings.warn('Start time is None and default start time is not set, '
                      'onetick.query will use 19700101 as start time, '
                      'which can cause unexpected results. '
                      'Please set start time explicitly.')
    if (end is None or end is utils.adaptive) and otp.config.get('default_end_time') is None and \
            not has_source_end:
        warnings.warn('End time is None and default end time is not set, '
                      'onetick.query will use 19700101 as end time, '
                      'which can cause unexpected results. '
                      'Please set end time explicitly.')

    if isinstance(start, _Operation) and start_time_expression is None:
        start_time_expression = str(start)
    if isinstance(end, _Operation) and end_time_expression is None:
        end_time_expression = str(end)

    if isinstance(start_time_expression, _Operation):
        start_time_expression = str(start_time_expression)
    if isinstance(end_time_expression, _Operation):
        end_time_expression = str(end_time_expression)

    # PY-1321: CEP-query seems to be using start and end values for some reason, so setting them to None
    if start_time_expression is not None:
        start = None
    if end_time_expression is not None:
        end = None

    if inspect.ismethod(query) or inspect.isfunction(query):
        t_s = None
        if isinstance(symbols, otp.Source):
            t_s = symbols
        if isinstance(symbols, otp.query):
            t_s = otp.Query(symbols)
        if isinstance(symbols, str):
            t_s = otp.Tick(SYMBOL_NAME=symbols)
        if isinstance(symbols, list):
            t_s = otp.Ticks(SYMBOL_NAME=symbols)

        if isinstance(t_s, otp.Source):
            query = query(t_s.to_symbol_param())  # type: ignore

    query, query_params = _preprocess_otp_query(query, query_params)
    # If query is an otp.Source object, then it can deal with otp.datetime and pd.Timestamp types

    if log_symbol is utils.default:
        log_symbol = otp.config.log_symbol
    if callback is None and log_symbol:
        callback = LogCallback(query)

    if manual_dataframe_callback:
        if output_structure and output_structure != 'df':
            raise ValueError("Parameter 'output_structure' must be set to 'df'"
                             " if parameter 'manual_dataframe_callback' is set")
        if log_symbol:
            raise ValueError("Parameters 'manual_dataframe_callback' and 'log_symbol' can't be set together")
        if callback is not None:
            raise ValueError("Parameters 'manual_dataframe_callback' and 'callback' can't be set together")
        callback = ManualDataframeCallback(timezone)

    output_mode = otq.QueryOutputMode.numpy
    if callback is not None:
        output_mode = otq.QueryOutputMode.callback
    if output_structure == 'polars':
        if not otq.webapi:
            raise ValueError("Parameter output_structure='polars' is only supported in WebAPI mode.")
        try:
            import polars as _  # type: ignore
        except ImportError:
            raise ValueError("Parameter output_structure='polars' is specified, but module polars can't be imported. "
                             "Use 'pip install onetick-py[polars]' command to install onetick-py with polars support.")
        try:
            output_mode = otq.QueryOutputMode.polars
        except AttributeError:
            raise ValueError("Parameter output_structure='polars' is specified, but it's not supported "
                             "by installed onetick.query_webapi library.")

    output_structure, output_structure_for_otq = _process_output_structure(output_structure)

    require_dict = require_dict or _is_dict_required(symbols)

    # converting symbols properly
    if isinstance(symbols, otp.Source):
        # check if SYMBOL_NAME is in schema, or if schema contains only one field
        if ('SYMBOL_NAME' not in symbols.columns(skip_meta_fields=True).keys()) and \
                len(symbols.columns(skip_meta_fields=True)) != 1:
            warnings.warn('Using as a symbol list a source without "SYMBOL_NAME" field '
                          'and with more than one field! This won\'t work unless the schema is incomplete')

        symbols = otp.Source._convert_symbol_to_string(
            symbol=symbols,
            tmp_otq=query._tmp_otq if isinstance(query, otp.Source) else None,
            start=start,
            end=end,
            timezone=timezone
        )
    if isinstance(symbols, str):
        symbols = [symbols]
    if isinstance(symbols, pd.DataFrame):
        symbols = utils.get_symbol_list_from_df(symbols)

    if isinstance(query, dict):
        # we assume it's a dictionary of sources for the MultiOutputSource object
        query = otp.MultiOutputSource(query)

    if symbol_date:
        # otq.run supports only strings and datetime.date
        symbol_date = utils.symbol_date_to_str(symbol_date)

    params_saved_to_otq = {}
    if isinstance(query, (otp.Source, otp.MultiOutputSource)):
        start = None if start is utils.adaptive else start
        end = None if end is utils.adaptive else end
        params_saved_to_otq = dict(
            symbols=symbols,
            start=start,
            end=end,
            start_time_expression=start_time_expression,
            end_time_expression=end_time_expression,
        )
        param_upd = query._prepare_for_execution(symbols=symbols, start=start, end=end,
                                                 timezone=timezone,
                                                 start_time_expression=start_time_expression,
                                                 end_time_expression=end_time_expression,
                                                 require_dict=require_dict,
                                                 running_query_flag=running,
                                                 node_name=node_name, has_output=None,
                                                 symbol_date=symbol_date)
        query, require_dict, node_name, symbol_date_to_run = param_upd
        # symbols and start/end times should be already stored in the query and should not be passed again
        symbols = None
        start = None
        end = None
        start_time_expression = None
        end_time_expression = None
        time_as_nsec = True
        # PY-1423: except for symbol date
        symbol_date = symbol_date_to_run

    elif isinstance(query, (otq.graph_components.EpBase, otq.Chainlet)):
        query = otq.GraphQuery(query)

    if isinstance(query, otq.SqlQuery):
        # This has no impact on query result, just placeholder values
        start = end = None

    if start is utils.adaptive:
        start = configuration.config.default_start_time

    if end is utils.adaptive:
        end = configuration.config.default_end_time

    if not otq.webapi:
        # converting to expressions, because in datetime objects nanoseconds are not supported on some OneTick versions
        if start is not None and not start_time_expression:
            start_time_expression = datetime2expr(start)
        if end is not None and not end_time_expression:
            end_time_expression = datetime2expr(end)

    # start and end parameters could be set to None,
    # because we use start and end time expressions,
    # but because of the bug it sometimes doesn't work
    # https://onemarketdata.atlassian.net/browse/BDS-454
    start, end = _get_start_end(start, end, timezone)

    # authentication
    username = username or otp.config.default_username
    alternative_username = alternative_username or otp.config.default_auth_username
    password = password or otp.config.default_password
    kwargs = {}
    if password is not None and has_password_param(throw_warning=True):
        kwargs['password'] = password

    max_expected_ticks_per_symbol = max_expected_ticks_per_symbol or otp.config.max_expected_ticks_per_symbol
    if max_expected_ticks_per_symbol is not None and has_max_expected_ticks_per_symbol(throw_warning=True):
        kwargs['max_expected_ticks_per_symbol'] = max_expected_ticks_per_symbol
    elif max_expected_ticks_per_symbol is None and has_max_expected_ticks_per_symbol(throw_warning=False):
        kwargs['max_expected_ticks_per_symbol'] = 2000

    if encoding is not None and has_query_encoding_parameter(throw_warning=True):
        kwargs['encoding'] = encoding

    run_params = dict(
        query=query,
        symbols=symbols, start=start, end=end, context=context, username=username,
        timezone=timezone,
        start_time_expression=start_time_expression,
        end_time_expression=end_time_expression,
        alternative_username=alternative_username, batch_size=batch_size,
        running_query_flag=running, query_properties=query_properties,
        max_concurrency=concurrency, apply_times_daily=apply_times_daily, symbol_date=symbol_date,
        query_params=query_params, time_as_nsec=time_as_nsec,
        treat_byte_arrays_as_strings=treat_byte_arrays_as_strings,
        output_mode=output_mode,
        output_matrix_per_field=output_matrix_per_field, output_structure=output_structure_for_otq,
        return_utc_times=return_utc_times, connection=connection,
        callback=callback, svg_path=svg_path, use_connection_pool=use_connection_pool, **kwargs
    )

    # some parameters were saved in .otq file, we need to debug them too
    debug_params = dict(run_params, **params_saved_to_otq) if params_saved_to_otq else run_params
    otp.get_logger(__name__).info(otp.utils.json_dumps(debug_params))

    try:
        result = otq.run(**run_params)
    except Exception as e:
        e = _add_stack_info_to_exception(e)
        e = _add_version_info_to_exception(e)
        raise e  # noqa: W0707

    if output_mode == otq.QueryOutputMode.callback:
        if manual_dataframe_callback:
            result = callback.result
        return result

    # node_names should be either a list of node names or None
    node_names: Optional[List[str]]
    if isinstance(node_name, str):
        node_names = [node_name]
    else:
        node_names = node_name

    if query_schema:
        # check if we have empty result for any symbol to add schema to empty dataframes
        _process_empty_results(result, query_schema, output_structure)

    return _format_call_output(result, output_structure=output_structure,
                               require_dict=require_dict, node_names=node_names,
                               print_symbol_errors=print_symbol_errors)


async def run_async(*args, **kwargs):
    """
    Asynchronous alternative to :func:`otp.run <onetick.py.run>`.

    All parameters are the same.

    This function can be used via built-in python ``await`` syntax
    and standard `asyncio <https://docs.python.org/3/library/asyncio.html>`_ library.

    Note
    ----
    Internally this function is implemented as :func:`otp.run <onetick.py.run>` running in a separate thread.

    Threads in python are generally not interruptable,
    so some `asyncio` functionality may not work as expected.

    For example, canceling :func:`otp.run_async <onetick.py.run_async>` task by
    `timeout <https://docs.python.org/3/library/asyncio-task.html#timeouts>`_
    may block the waiting function
    or exiting the python process will block until the task is finished,
    depending on python and `asyncio` back-end implementation.

    Examples
    --------

    >>> data = otp.Ticks(A=[1, 2, 3])

    Calling :func:`otp.run_async <onetick.py.run_async>` will create a "coroutine" object:

    >>> otp.run_async(data) # doctest: +SKIP
    <coroutine object run_async at ...>

    Use `asyncio.run <https://docs.python.org/3/library/asyncio-runner.html#asyncio.run>`_
    to run this coroutine and wait for it to finish:

    >>> asyncio.run(otp.run_async(data))
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3

    You can use standard `asyncio <https://docs.python.org/3/library/asyncio.html>`_
    library functions to create and schedule tasks.

    In the example below two tasks are executed in parallel,
    so total execution time will be around 3 seconds instead of 6:

    >>> import asyncio
    >>> import time
    >>> async def parallel_tasks():
    ...     # pause 1 second on each tick (thus 3 seconds for 3 ticks)
    ...     task_otp = asyncio.create_task(otp.run_async(data.pause(1000)))
    ...     # just sleep for 3 seconds
    ...     task_other = asyncio.create_task(asyncio.sleep(3))
    ...     otp_result = await task_otp
    ...     await task_other
    ...     print(otp_result)
    >>> start_time = time.time()
    >>> asyncio.run(parallel_tasks())
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3
    >>> print('Finished in', time.time() - start_time, 'seconds') # doctest: +SKIP
    Finished in 3.0108885765075684 seconds
    """
    return await asyncio.to_thread(run, *args, **kwargs)


def _filter_returned_map_by_node(result, _node_names):
    """
    Here, result has the following format: {symbol: {node_name: data}}
    We need to filter by correct node_name
    """
    # TODO: implement filtering by node_name in a way
    # that no information from SymbolNumpyResultMap object is lost
    return result


def _filter_returned_list_by_node(result, node_names):
    """
    Here, result has the following format: [(symbol, ticks_data, error_data, node_name)]
    We need to filter by correct node_names
    """
    if not node_names:
        return result

    node_found = False

    res = []
    empty_result = True
    for symbol, ticks_data, error_data, node, *_ in result:
        if len(ticks_data) > 0:
            empty_result = False
        if node in node_names:
            node_found = True
            res.append((symbol, ticks_data, error_data, node))

    if not empty_result and not node_found:
        # TODO: Do we even want to raise it?
        raise ValueError(f'No passed node name(s) were found in the results. Passed node names were: {node_names}')
    return res


def _form_dict_from_list(data_list, output_structure, print_symbol_errors):
    """
    Here, data_list has the following format: [(symbol, ticks_data, error_data, node_name), ...]
    We need to create the following result:
    either {symbol: DataFrame(ticks_data)} if there is only one result per symbol
    or {symbol: [DataFrame(ticks_data)]} if there are multiple results for symbol for a single node_name
    or {symbol: {node_name: DataFrame(ticks_data)}} if there are single results for multiple node names for a symbol
    or {symbol: {node_name: [DataFrame(ticks_data)]}} if there are multiple results for multiple node names for a symbol
    """

    def form_node_name_dict(lst):
        """
        lst is a list of (node, dataframe)
        """
        d = defaultdict(list)
        for node, df in lst:
            d[node].append(df)
        for node, node_list in d.items():
            if len(node_list) == 1:
                d[node] = node_list[0]
        if len(d) == 1:
            d = list(d.values())[0]
        else:  # converting defaultdict to regular dict
            d = dict(d)
        return d

    def get_dataframe(data):
        if output_structure == 'df':
            return pd.DataFrame(dict(data))
        else:
            import polars
            if isinstance(data, polars.DataFrame):
                # polars only works in webapi mode,
                # and it's already returned as polars.DataFrame by onetick.query_webapi
                return data
            # but if there is no data, then we want to return empty polars.DataFrame
            return polars.DataFrame()

    symbols_dict = defaultdict(list)
    for symbol, data, error_data, node, *_ in data_list:

        if print_symbol_errors:
            for err_code, err_msg, *_ in error_data:
                warnings.warn(f"Symbol error: [{err_code}] {err_msg}")

        df = get_dataframe(data)

        list_item = (node, df)
        symbols_dict[symbol].append(list_item)

    for symbol, lst in symbols_dict.items():
        symbols_dict[symbol] = form_node_name_dict(lst)

    return dict(symbols_dict)


def _format_call_output(result, output_structure, node_names, require_dict, print_symbol_errors):
    """Formats output of otq.run() according to passed parameters.
    See parameters' description for more information

    Parameters
    ----------
    output_structure: ['df', 'list', 'map']
        If 'df': forms pandas.DataFrame from the result.

        Returns a dictionary with symbols as keys if there's more than one symbol
        in returned data of if require_dict = True.

        Values of the returned dictionary, or returned value itself if no dictionary is formed,
        is either a list of tuples: (node_name, dataframe) if there's output for more than one node
        or a dataframe

        If 'list' or 'map': returns data as returned by otq.run(), possibly filtered by node_name (see below)
    node_names: str, None
        If not None, then selects only output returned by nodes in node_names list
        for all output structures
    require_dict: bool
        If True, forces output for output_structure='df' to always be a dictionary, even if only one symbol is returned
        Has no effect for other values of output_structure
    print_symbol_errors: bool
        Print OneTick symbol errors in when ``output_structure`` is set to 'df' or not.

    Returns
    ----------
        Formatted output: pandas DataFrame, dictionary or list

    """
    if output_structure == 'list':
        return _filter_returned_list_by_node(result, node_names)
    elif output_structure == 'map':
        return _filter_returned_map_by_node(result, node_names)

    assert output_structure in ('df', 'polars'), (f'Output structure should be one of: "df", "map", "list", "polars" '
                                                  f'instead "{output_structure}" was passed')

    # "df" output structure implies that raw results came as a list
    result_list = _filter_returned_list_by_node(result, node_names)
    result_dict = _form_dict_from_list(result_list, output_structure, print_symbol_errors)

    if len(result_dict) == 1 and not require_dict:
        return list(result_dict.values())[0]
    else:
        return result_dict


def _process_empty_results(result, query_schema, output_structure):
    """
    Process query results and add columns to empty responses based on query schema.
    """
    schema = [
        (field, np.array([], dtype=otp.types.type2np(dtype)))
        for field, dtype in {**query_schema, 'Time': otp.nsectime}.items()
    ]
    if isinstance(result, otq.SymbolNumpyResultMap):
        empty_data = dict(schema)
    else:
        empty_data = schema

    if output_structure == 'polars':
        import polars
        empty_data = polars.DataFrame(dict(schema))

    if isinstance(result, otq.SymbolNumpyResultMap):
        for result_item in result.get_dict().values():
            for node_name, symbol_result in result_item.items():
                if len(symbol_result[0]) == 0:
                    result_item[node_name] = (empty_data, symbol_result[1])
    else:
        for idx, result_item in enumerate(result):
            if len(result_item[1]) == 0:
                result[idx] = (
                    result_item[0], empty_data, result_item[2], result_item[3], *result_item[4:]
                )


def _preprocess_otp_query(query, query_params):

    if isinstance(query, otp.query._outputs):
        query = query['OUT']

    if isinstance(query, otp.query):
        if query.params:
            if query_params:
                raise ValueError("please specify parameters in query or in otp.run only")
            query_params = query.params
        query = query.path
    return query, query_params


def _get_start_end(start, end, timezone):
    """
    Convert datetime objects supported by onetick-py
    to datetime objects supported by onetick-query.
    """
    def support_nanoseconds(time):
        if isinstance(time, (pd.Timestamp, otp.datetime)):
            if otq.webapi:
                # onetick-query_webapi supports pandas.Timestamp and strings in %Y%m%s%H%M%S.%J format
                if isinstance(time, pd.Timestamp):
                    return time
                elif isinstance(time, otp.datetime):
                    return time.ts
            else:
                if otp.compatibility.is_correct_timezone_used_in_otq_run():
                    time = datetime2timeval(time, timezone)
                else:
                    # there is a bug in older onetick versions using wrong timezone
                    time = datetime2timeval(time, 'GMT')
        return time

    if start is utils.adaptive:
        start = configuration.config.default_start_time

    if end is utils.adaptive:
        end = configuration.config.default_end_time

    # `isinstance(obj, datetime.date)` is not correct because
    # isinstance(<datetime.datetime object>, datetime.date) = True
    # pylint: disable=unidiomatic-typecheck
    if type(start) is datetime.date:
        start = datetime.datetime(start.year, start.month, start.day)
    if type(end) is datetime.date:
        end = datetime.datetime(end.year, end.month, end.day)

    start = support_nanoseconds(start)
    end = support_nanoseconds(end)

    return start, end


def _process_output_structure(output_structure):
    if not output_structure or output_structure == "df":  # otq doesn't support df
        output_structure = "df"
        output_structure_for_otq = "symbol_result_list"
    elif output_structure == "list":
        output_structure_for_otq = "symbol_result_list"
    elif output_structure == "map":
        output_structure_for_otq = "symbol_result_map"
    elif output_structure == "polars":
        output_structure = "polars"
        output_structure_for_otq = "symbol_result_list"
    else:
        raise ValueError("output_structure support only the following values: df, list, map and polars")
    return output_structure, output_structure_for_otq



import datetime as dt
import inspect
import warnings

from typing import Dict, Iterable, Optional

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.db import _inspection
from onetick.py.core._source._symbol_param import _SymbolParamColumn
from onetick.py.core._source.tmp_otq import TmpOtq
from onetick.py.core.eval_query import _QueryEvalWrapper
from onetick.py.core.source import Source
from onetick.py.core.column_operations.base import Raw, OnetickParameter

from .. import types as ott
from .. import utils
from ..core.column_operations.base import _Operation
from ..db.db import DB
from ..compatibility import is_supported_where_clause_for_back_ticks

from onetick.py.docs.utils import docstring, param_doc

from .common import convert_tick_type_to_str, get_start_end_by_date
from .symbols import Symbols
from .ticks import Ticks
from .query import query


_db_doc = param_doc(
    name='db',
    desc="""
    Name(s) of the database or the database object(s).
    """,
    str_annotation='str, list of str, :class:`otp.DB <onetick.py.DB>`',
    default=None,
    str_default='None',
)
_symbol_doc = param_doc(
    name='symbol',
    desc="""
    Symbol(s) from which data should be taken.
    """,
    str_annotation='str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`',
    default=utils.adaptive,
    str_default=' :py:class:`onetick.py.adaptive`',
)
_symbols_doc = param_doc(
    name='symbols',
    desc="""
    Symbol(s) from which data should be taken.
    Alias for ``symbol`` parameter. Will take precedence over it.
    """,
    str_annotation=('str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`, '
                    ':py:class:`onetick.query.GraphQuery`.'),
    default=None,
)
_tick_type_doc = param_doc(
    name='tick_type',
    desc="""
    Tick type of the data.
    If not specified, all ticks from `db` will be taken.
    If ticks can't be found or there are many databases specified in `db` then default is "TRD".
    """,
    str_annotation='str, list of str',
    default=utils.adaptive,
    str_default=' :py:class:`onetick.py.adaptive`',
)
_start_doc = param_doc(
    name='start',
    desc="""
    Start of the interval from which the data should be taken.
    Default is :py:class:`onetick.py.adaptive`, making the final query deduce the time
    limits from the rest of the graph.
    """,
    str_annotation=(
        ':py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`,'
        ' :py:class:`onetick.py.adaptive`'
    ),
    default=utils.adaptive,
    str_default=' :py:class:`onetick.py.adaptive`',
)
_end_doc = param_doc(
    name='end',
    desc="""
    End of the interval from which the data should be taken.
    Default is :py:class:`onetick.py.adaptive`, making the final query deduce the time
    limits from the rest of the graph.
    """,
    str_annotation=(
        ':py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`,'
        ' :py:class:`onetick.py.adaptive`'
    ),
    default=utils.adaptive,
    str_default=' :py:class:`onetick.py.adaptive`',
)
_date_doc = param_doc(
    name='date',
    desc="""
    Allows to specify a whole day instead of passing explicitly ``start`` and ``end`` parameters.
    If it is set along with the ``start`` and ``end`` parameters then last two are ignored.
    """,
    str_annotation=":class:`datetime.datetime`, :class:`otp.datetime <onetick.py.datetime>`",
    default=None,
)
_schema_policy_doc = param_doc(
    name='schema_policy',
    desc="""
    Schema deduction policy:

    - 'tolerant' (default)
      The resulting schema is a combination of ``schema`` and database schema.
      If the database schema can be deduced,
      it's checked to be type-compatible with a ``schema``,
      and ValueError is raised if checks are failed.
      Also, with this policy database is scanned 5 days back to find the schema.
      It is useful when database is misconfigured or in case of holidays.

    - 'tolerant_strict'
      The resulting schema will be ``schema`` if it's not empty.
      Otherwise, database schema is used.
      If the database schema can be deduced,
      it's checked if it lacks fields from the ``schema``
      and it's checked to be type-compatible with a ``schema``
      and ValueError is raised if checks are failed.
      Also, with this policy database is scanned 5 days back to find the schema.
      It is useful when database is misconfigured or in case of holidays.

    - 'fail'
      The same as 'tolerant', but if the database schema can't be deduced, raises an Exception.

    - 'fail_strict'
      The same as 'tolerant_strict', but if the database schema can't be deduced, raises an Exception.

    - 'manual'
      The resulting schema is a combination of ``schema`` and database schema.
      Compatibility with database schema will not be checked.

    - 'manual_strict'
      The resulting schema will be exactly ``schema``.
      Compatibility with database schema will not be checked.
      If some fields specified in ``schema`` do not exist in the database,
      their values will be set to some default value for a type
      (0 for integers, NaNs for floats, empty string for strings, epoch for datetimes).

    Default value is :py:class:`onetick.py.adaptive` (if deprecated parameter ``guess_schema`` is not set).
    If ``guess_schema`` is set to True then value is 'fail', if False then 'manual'.
    If ``schema_policy`` is set to ``None`` then default value is 'tolerant'.

    Default value can be changed with
    :py:attr:`otp.config.default_schema_policy<onetick.py.configuration.Config.default_schema_policy>`
    configuration parameter.

    If you set schema manually, while creating DataSource instance, and don't set ``schema_policy``,
    it will be automatically set to ``manual``.
    """,
    str_annotation="'tolerant', 'tolerant_strict', 'fail', 'fail_strict', 'manual', 'manual_strict'",
    default=utils.adaptive,
    str_default=' :py:class:`onetick.py.adaptive`',
)
_guess_schema_doc = param_doc(
    name='guess_schema',
    desc="""
    .. deprecated:: 1.3.16

    Use ``schema_policy`` parameter instead.

    If ``guess_schema`` is set to True then ``schema_policy`` value is 'fail', if False then 'manual'.
    """,
    annotation=bool,
    default=None,
)
_identify_input_ts_doc = param_doc(
    name='identify_input_ts',
    desc="""
    If set to False, the fields SYMBOL_NAME and TICK_TYPE are not appended to the output ticks.
    """,
    annotation=bool,
    default=False,
)
_back_to_first_tick_doc = param_doc(
    name='back_to_first_tick',
    desc="""
    Determines how far back to go looking for the latest tick before ``start`` time.
    If one is found, it is inserted into the output time series with the timestamp set to ``start`` time.
    Note: it will be rounded to int, so otp.Millis(999) will be 0 seconds.
    """,
    str_annotation=('int, :ref:`offset <datetime_offsets>`, '
                    ':class:`otp.expr <onetick.py.expr>`, '
                    ':py:class:`~onetick.py.Operation`'),
    default=0,
)
_keep_first_tick_timestamp_doc = param_doc(
    name='keep_first_tick_timestamp',
    desc="""
    If set, new field with this name will be added to source.
    This field contains original timestamp of the tick that was taken from before the start time of the query.
    For all other ticks value in this field will be equal to the value of Time field.
    This parameter is ignored if ``back_to_first_tick`` is not set.
    """,
    annotation=str,
    default=None,
)
_presort_doc = param_doc(
    name='presort',
    desc="""
    Add the presort EP in case of bound symbols.
    Applicable only when ``symbols`` is not None.
    By default, it is set to True if ``symbols`` are set
    and to False otherwise.
    """,
    annotation=bool,
    default=utils.adaptive,
    str_default=' :py:class:`onetick.py.adaptive`',
)
_concurrency_doc = param_doc(
    name='concurrency',
    desc="""
    Specifies the number of CPU cores to utilize for the ``presort``.
    By default, the value is inherited from the value of the query where this PRESORT is used.

    For the main query it may be specified in the ``concurrency`` parameter of :meth:`run` method
    (which by default is set to
    :py:attr:`otp.config.default_concurrency<onetick.py.configuration.Config.default_concurrency>`).

    For the auxiliary queries (like first-stage queries) empty value means OneTick's default of 1.
    If :py:attr:`otp.config.presort_force_default_concurrency<onetick.py.configuration.Config.presort_force_default_concurrency>`
    is set then default concurrency value will be set in all PRESORT EPs in all queries.
    """,  # noqa: E501
    annotation=int,
    default=utils.default,
    str_default=' :py:class:`onetick.py.utils.default`',
)
_batch_size_doc = param_doc(
    name='batch_size',
    desc="""
    Specifies the query batch size for the ``presort``.
    By default, the value from
    :py:attr:`otp.config.default_batch_size<onetick.py.configuration.Config.default_batch_size>` is used.
    """,
    annotation=int,
    default=None,
)
_schema_doc = param_doc(
    name='schema',
    desc="""
    Dict of <column name> -> <column type> pairs that the source is expected to have.
    If the type is irrelevant, provide None as the type in question.
    """,
    annotation=Optional[Dict[str, type]],
    default=None,
)
_desired_schema_doc = param_doc(
    name='kwargs',
    desc="""
    Deprecated. Use ``schema`` instead.
    List of <column name> -> <column type> pairs that the source is expected to have.
    If the type is irrelevant, provide None as the type in question.
    """,
    str_annotation='type[str]',
    kind=inspect.Parameter.VAR_KEYWORD,
)

_max_back_ticks_to_prepend_doc = param_doc(
    name='max_back_ticks_to_prepend',
    desc="""
    When the ``back_to_first_tick`` interval is specified, this parameter determines the maximum number
    of the most recent ticks before start_time that will be prepended to the output time series.
    Their timestamp will be changed to start_time.
    """,
    annotation=int,
    default=1,
)

_where_clause_for_back_ticks_doc = param_doc(
    name='where_clause_for_back_ticks',
    desc="""
    A logical expression that is computed only for the ticks encountered when a query goes back from the start time,
    in search of the ticks to prepend. If it returns false, a tick is ignored.
    """,
    annotation=Raw,
    default=None,
)
_symbol_date_doc = param_doc(
    name='symbol_date',
    desc="""
    Symbol date or integer in the YYYYMMDD format.
    Can only be specified if parameters ``symbols`` is set.
    """,
    str_annotation=':py:class:`otp.datetime <onetick.py.datetime>` or :py:class:`datetime.datetime` or int',
    default=None,
)

DATA_SOURCE_DOC_PARAMS = [
    _db_doc, _symbol_doc, _tick_type_doc,
    _start_doc, _end_doc, _date_doc,
    _schema_policy_doc, _guess_schema_doc,
    _identify_input_ts_doc,
    _back_to_first_tick_doc, _keep_first_tick_timestamp_doc,
    _max_back_ticks_to_prepend_doc,
    _where_clause_for_back_ticks_doc,
    _symbols_doc,
    _presort_doc, _batch_size_doc, _concurrency_doc,
    _schema_doc,
    _symbol_date_doc,
    _desired_schema_doc,
]


class DataSource(Source):

    POLICY_MANUAL = "manual"
    POLICY_MANUAL_STRICT = "manual_strict"
    POLICY_TOLERANT = "tolerant"
    POLICY_TOLERANT_STRICT = "tolerant_strict"
    POLICY_FAIL = "fail"
    POLICY_FAIL_STRICT = "fail_strict"

    _VALID_POLICIES = frozenset([POLICY_MANUAL, POLICY_MANUAL_STRICT,
                                 POLICY_TOLERANT, POLICY_TOLERANT_STRICT,
                                 POLICY_FAIL, POLICY_FAIL_STRICT])
    _PROPERTIES = Source._PROPERTIES + ["_p_db", "_p_strict", "_p_schema", "_schema", "logger"]

    def __get_schema(self, db, start, schema_policy):
        schema = {}

        if start is utils.adaptive:
            start = None  # means that use the last date with data

        if isinstance(db, list):
            # This case of a merge, since we need to get combined schema across different tick types and dbs
            for t_db in db:
                if t_db.startswith('expr('):
                    continue

                _db = t_db.split(':')[0]
                _tt = t_db.split(':')[-1]

                # tick type as parameter
                if _tt.startswith('$'):
                    _tt = None

                db_obj = _inspection.DB(_db)
                if schema_policy == self.POLICY_TOLERANT and start:
                    # repeating the same logic as in db_obj.last_date
                    start = db_obj.last_not_empty_date(start, days_back=5, tick_type=_tt)

                db_schema = {}
                try:
                    db_schema = db_obj.schema(date=start, tick_type=_tt)
                except Exception as e:
                    if _tt is not None:
                        warnings.warn(f"Couldn't get schema from the database {db_obj}:\n{e}.\n\n"
                                      "Set parameter schema_policy='manual' to set the schema manually.")

                schema.update(db_schema)

        if db is None or isinstance(db, _SymbolParamColumn):
            # In this case we can't get schema, because db is calculated dynamically.
            # Set to empty to indicate that in this case we expect the manually set schema.
            schema = {}
        return schema

    def __prepare_schema(self, db, start, schema_policy, guess_schema, schema):
        if guess_schema is not None:
            warnings.warn(
                "guess_schema flag is deprecated; use schema_policy argument instead",
                FutureWarning,
            )
            if schema_policy is not None:
                raise ValueError("guess_schema and schema_policy cannot be set at the same time")
            if guess_schema:
                schema_policy = self.POLICY_FAIL
            else:
                schema_policy = self.POLICY_MANUAL

        if schema_policy is None:
            schema_policy = self.POLICY_TOLERANT
        if schema_policy not in self._VALID_POLICIES:
            raise ValueError(f"Invalid schema_policy; allowed values are: {self._VALID_POLICIES}")

        actual_schema = {}
        if schema_policy not in (self.POLICY_MANUAL, self.POLICY_MANUAL_STRICT):
            actual_schema = self.__get_schema(db, start, schema_policy)
            dbs = ', '.join(db if isinstance(db, list) else [])

            if len(actual_schema) == 0:
                if schema_policy in (self.POLICY_FAIL, self.POLICY_FAIL_STRICT):
                    raise ValueError(f'No ticks found in database(-s) {dbs}')
                # lets try to use at least something
                return schema.copy()

            for k, v in schema.items():
                field_type = actual_schema.get(k, None)
                incompatible_types = False
                if field_type is None:
                    if self._p_strict or schema_policy in (self.POLICY_TOLERANT, self.POLICY_FAIL):
                        raise ValueError(f"Database(-s) {dbs} schema has no {k} field")
                elif issubclass(field_type, str) and issubclass(v, str):
                    field_length = ott.string.DEFAULT_LENGTH
                    if issubclass(field_type, ott.string):
                        field_length = field_type.length
                    v_length = ott.string.DEFAULT_LENGTH
                    if issubclass(v, ott.string):
                        v_length = v.length
                    if issubclass(field_type, ott.varstring):
                        if not issubclass(v, ott.varstring):
                            incompatible_types = True
                    elif not issubclass(v, ott.varstring) and v_length < field_length:
                        incompatible_types = True
                elif not issubclass(field_type, v):
                    incompatible_types = True
                if incompatible_types:
                    error_message = f"Database(-s) {dbs} schema field {k} has type {field_type}, but {v} was requested"
                    if field_type in (str, ott.string) or v in (str, ott.string):
                        error_message = f"{error_message}. Notice, that `str` and `otp.string` lengths are 64"
                    raise ValueError(error_message)
            if not self._p_strict:
                schema.update(actual_schema)

        table_schema = schema.copy()
        if not self._p_strict:
            # in this case we will table only fields specified by user
            table_schema = {
                k: v for k, v in table_schema.items() if k not in actual_schema
            }
        return table_schema

    def __prepare_db_tick_type(self, db, tick_type, start, end):
        if isinstance(db, list):
            # If everything is correct then this branch should leave
            # the `db` var as a list of databases with tick types and the `tick_type` var is None.
            # Valid cases:
            #     - Fully defined case. The `db` parameter has a list of databases where
            #       every database has a tick type, when the `tick_type`
            #       parameter has default value or None (for backward compatibility)
            #     - Partially defined case. The `db` parameter has a list of databases but
            #       not every database has a tick type, and meantime the `tick_type`
            #       is passed to not None value. In that case databases without tick type
            #       are exetended with a tick type from the `tick_type` parameter
            #     - No defined case. The `db` parameter has a list of databases and
            #       every database there has no tick type, and the `tick_type` is
            #       set to not None value. In that case every database is extended with
            #       the tick type from the `tick_type`.

            def db_converter(_db):
                if isinstance(_db, DB):
                    return _db.name
                else:
                    return _db

            db = [db_converter(_db) for _db in db]
            res = all(('::' in _db and _db[-1] != ':' for _db in db))
            if res:
                if tick_type is utils.adaptive or tick_type is None:
                    tick_type = None  # tick types is specified for all databases
                else:
                    raise ValueError('The `tick_type` is set as a parameter '
                                     'and also as a part of the `db` parameter'
                                     'for every database')
            else:
                dbs_without_tt = [_db.split(':')[0] for _db in db
                                  if '::' not in _db or _db[-1] == ':']

                if tick_type is utils.adaptive:
                    tick_type = 'TRD'  # default one for backward compatibility and testing usecase
                if tick_type is None:
                    raise ValueError('The tick type is not set for databases: ' +
                                     ', '.join(dbs_without_tt))
                else:
                    # extend databases with missing tick types from the tick tick parameter
                    dbs_with_tt = [_db for _db in db
                                   if '::' in _db and _db[-1] != ':']

                    db = dbs_with_tt + [_db + '::' + tick_type for _db in dbs_without_tt]
                    tick_type = None

        if isinstance(db, (DB, _inspection.DB)):
            db = db.name  # ... and we go to the next branch

        if isinstance(db, str):
            # The resulting `db` var contains a list with string value, that has the `db`
            # concatenated with the `tick_type`.
            if '::' in db:
                if tick_type is utils.adaptive or tick_type is None:
                    tick_type = db.split(':')[-1]
                    db = db.split('::')[0]
                else:
                    raise ValueError('The `tick_type` is set as a parameter '
                                     'and also as a part of the `db` parameter')
            else:
                if tick_type is utils.adaptive or tick_type is None:
                    db_obj = _inspection.DB(db)

                    # try to find at least one common tick type
                    # through all days
                    tick_types = None

                    if start is utils.adaptive:
                        start = end = db_obj.get_last_date(show_warnings=False)

                    if start and end:  # could be None if there is no data
                        t_start = start
                        while t_start <= end:
                            t_tts = set(db_obj.tick_types(t_start))

                            t_start += dt.timedelta(days=1)

                            if len(t_tts) == 0:
                                continue

                            if tick_types is None:
                                tick_types = t_tts
                            else:
                                tick_types &= t_tts

                            if len(tick_types) == 0:
                                raise ValueError(f'It seems that there is no common '
                                                 f'tick types for dates from {start} '
                                                 f'to {end}. Please specify a tick '
                                                 'type')

                    if tick_types is None:
                        if tick_type is utils.adaptive:
                            tick_types = ['TRD']  # the default one
                        else:
                            raise ValueError(f'Could not find any data in from {start} '
                                             f' to {end}. Could you check that tick type, '
                                             ' database and date range are correct.')

                    if len(tick_types) != 1:
                        raise ValueError('The tick type is not specified, found '
                                         'multiple tick types in the database : ' +
                                         ', '.join(tick_types))

                    tick_type = tick_types.pop()

            if not isinstance(tick_type, str) and isinstance(tick_type, Iterable):
                if isinstance(tick_type, _SymbolParamColumn):
                    db = [f"expr('{db}::' + {str(tick_type)})"]
                else:
                    db = [f'{db}::{tt}' for tt in tick_type]
            else:
                db = [db + '::' + tick_type]
            tick_type = None

        if isinstance(db, _SymbolParamColumn):
            # Do nothing, because we don't know whether db will come with the tick type or not.
            # The only one thing that definetely we know that tick_type can not be utils.adaptive
            if tick_type is utils.adaptive:
                # TODO: need to test this case
                raise ValueError('The `db` is set to the symbol param, in that case '
                                 'the `tick_type` should be set explicitly to some value '
                                 'or to None')

        if db is None:
            # This case means that database comes with the symbol name, then tick type should be defined
            if tick_type is utils.adaptive or tick_type is None:
                raise ValueError('The `db` is not specified that means database is '
                                 'expected to be defined with the symbol name. '
                                 'In that case the `tick_type` should be defined.')
            if not isinstance(tick_type, str) and isinstance(tick_type, Iterable):
                tick_type = '+'.join(tick_type)

        return db, tick_type

    @docstring(parameters=DATA_SOURCE_DOC_PARAMS, add_self=True)
    def __init__(
        self,
        db=None,
        symbol=utils.adaptive,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        date=None,
        schema=None,
        schema_policy=utils.adaptive,
        guess_schema=None,
        identify_input_ts=False,
        back_to_first_tick=0,
        keep_first_tick_timestamp=None,
        max_back_ticks_to_prepend=1,
        where_clause_for_back_ticks=None,
        symbols=None,
        presort=utils.adaptive,
        batch_size=None,
        concurrency=utils.default,
        symbol_date=None,
        **kwargs,
    ):
        """
        Construct a source providing data from a given ``db``.

        .. warning::

            Default value of the parameter ``schema_policy`` enables automatic deduction
            of the data schema, but it is highly not recommended for production code.
            For details see :ref:`static/concepts/schema:Schema deduction mechanism`.

        Note
        ----
        If interval that was set for :py:class:`~onetick.py.DataSource` via ``start``/``end`` or ``date`` parameters
        does not match intervals in other :py:class:`~onetick.py.Source` objects used in query,
        or does not match the whole query interval, then :py:meth:`~otp.Source.modify_query_times` will be applied
        to this ``DataSource`` with specified interval as start and end times parameters.

        If ``symbols`` parameter is omitted, you need to specify unbound symbols for the query in ``symbols``
        parameter of :py:func:`onetick.py.run` function.

        If ``symbols`` parameter is set, :meth:`otp.merge <onetick.py.merge>` is used to merge all passed bound symbols.
        In this case you don't need to specify unbound symbols in :py:func:`onetick.py.run` call.

        It's not allowed to specify bound and unbound symbols at the same time.

        See also
        --------
        :ref:`static/concepts/start_end:Query start / end flow`
        :ref:`static/concepts/symbols:Symbols: bound and unbound`

        Examples
        --------

        Query a single symbol from a database:

        >>> data = otp.DataSource(db='SOME_DB', tick_type='TT', symbols='S1')
        >>> otp.run(data)
                             Time  X
        0 2003-12-01 00:00:00.000  1
        1 2003-12-01 00:00:00.001  2
        2 2003-12-01 00:00:00.002  3

        Parameter ``symbols`` can be a list.
        In this case specified symbols will be merged into a single data flow:

        >>> # OTdirective: snippet-name:fetch data.simple;
        >>> data = otp.DataSource(db='SOME_DB', tick_type='TT', symbols=['S1', 'S2'])
        >>> otp.run(data)
                             Time  X
        0 2003-12-01 00:00:00.000  1
        1 2003-12-01 00:00:00.000 -3
        2 2003-12-01 00:00:00.001  2
        3 2003-12-01 00:00:00.001 -2
        4 2003-12-01 00:00:00.002  3
        5 2003-12-01 00:00:00.002 -1

        Parameter ``identify_input_ts`` can be used to automatically add field with symbol name for each tick:

        >>> data = otp.DataSource(db='SOME_DB', tick_type='TT', symbols=['S1', 'S2'], identify_input_ts=True)
        >>> otp.run(data)
                             Time SYMBOL_NAME TICK_TYPE  X
        0 2003-12-01 00:00:00.000          S1        TT  1
        1 2003-12-01 00:00:00.000          S2        TT -3
        2 2003-12-01 00:00:00.001          S1        TT  2
        3 2003-12-01 00:00:00.001          S2        TT -2
        4 2003-12-01 00:00:00.002          S1        TT  3
        5 2003-12-01 00:00:00.002          S2        TT -1

        Source also can be passed as symbols, in such case magic named column SYMBOL_NAME will be transform to symbol
        and all other columns will be symbol parameters

        >>> # OTdirective: snippet-name:fetch data.symbols as a source;
        >>> symbols = otp.Ticks(SYMBOL_NAME=['S1', 'S2'])
        >>> data = otp.DataSource(db='SOME_DB', symbols=symbols, tick_type='TT')
        >>> otp.run(data)
                             Time  X
        0 2003-12-01 00:00:00.000  1
        1 2003-12-01 00:00:00.000 -3
        2 2003-12-01 00:00:00.001  2
        3 2003-12-01 00:00:00.001 -2
        4 2003-12-01 00:00:00.002  3
        5 2003-12-01 00:00:00.002 -1

        Default schema policy is **tolerant** (unless you specified ``schema`` parameter and
        left ``schema_policy`` with default value, when it will be set to **manual**).

        >>> data = otp.DataSource(
        ...     db='US_COMP', tick_type='TRD', symbols='AAPL', date=otp.dt(2022, 3, 1),
        ... )
        >>> data.schema
        {'PRICE': <class 'float'>, 'SIZE': <class 'int'>}

        >>> data = otp.DataSource(
        ...     db='US_COMP', tick_type='TRD', symbols='AAPL', schema={'PRICE': int},
        ...     schema_policy='tolerant', date=otp.dt(2022, 3, 1),
        ... )
        Traceback (most recent call last):
          ...
        ValueError: Database(-s) US_COMP::TRD schema field PRICE has type <class 'float'>,
        but <class 'int'> was requested

        Schema policy **manual** uses exactly ``schema``:

        >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL', schema={'PRICE': float},
        ...                       date=otp.dt(2022, 3, 1), schema_policy='manual')
        >>> data.schema
        {'PRICE': <class 'float'>}

        Schema policy **fail** raises an exception if the schema cannot be deduced:

        >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL', date=otp.dt(2021, 3, 1),
        ...                       schema_policy='fail')
        Traceback (most recent call last):
          ...
        ValueError: No ticks found in database(-s) US_COMP::TRD

        ``back_to_first_tick`` sets how far back to go looking for the latest tick before ``start`` time:

        >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL', date=otp.dt(2022, 3, 2),
        ...                       back_to_first_tick=otp.Day(1))
        >>> otp.run(data)
                             Time  PRICE  SIZE
        0 2022-03-02 00:00:00.000    1.4    50
        1 2022-03-02 00:00:00.000    1.0   100
        2 2022-03-02 00:00:00.001    1.1   101
        3 2022-03-02 00:00:00.002    1.2   102

        ``keep_first_tick_timestamp`` allows to show the original timestamp of the tick that was taken from before
        the start time of the query:

        >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL', date=otp.dt(2022, 3, 2),
        ...                       back_to_first_tick=otp.Day(1), keep_first_tick_timestamp='ORIGIN_TIMESTAMP')
        >>> otp.run(data)
                             Time         ORIGIN_TIMESTAMP  PRICE  SIZE
        0 2022-03-02 00:00:00.000  2022-03-01 00:00:00.002    1.4    50
        1 2022-03-02 00:00:00.000  2022-03-02 00:00:00.000    1.0   100
        2 2022-03-02 00:00:00.001  2022-03-02 00:00:00.001    1.1   101
        3 2022-03-02 00:00:00.002  2022-03-02 00:00:00.002    1.2   102

        ``max_back_ticks_to_prepend`` is used with ``back_to_first_tick``
        if more than 1 ticks before start time should be retrieved:

        >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL', date=otp.dt(2022, 3, 2),
        ...                       max_back_ticks_to_prepend=2, back_to_first_tick=otp.Day(1),
        ...                       keep_first_tick_timestamp='ORIGIN_TIMESTAMP')
        >>> otp.run(data)
                             Time         ORIGIN_TIMESTAMP  PRICE  SIZE
        0 2022-03-02 00:00:00.000  2022-03-01 00:00:00.001    1.4    10
        1 2022-03-02 00:00:00.000  2022-03-01 00:00:00.002    1.4    50
        2 2022-03-02 00:00:00.000  2022-03-02 00:00:00.000    1.0   100
        3 2022-03-02 00:00:00.001  2022-03-02 00:00:00.001    1.1   101
        4 2022-03-02 00:00:00.002  2022-03-02 00:00:00.002    1.2   102

        ``where_clause_for_back_ticks`` is used to filter out ticks before the start time:

        .. testcode::
           :skipif: not is_supported_where_clause_for_back_ticks()

           data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAPL', date=otp.dt(2022, 3, 2),
                                 where_clause_for_back_ticks=otp.raw('SIZE>=50', dtype=bool),
                                 back_to_first_tick=otp.Day(1), max_back_ticks_to_prepend=2,
                                 keep_first_tick_timestamp='ORIGIN_TIMESTAMP')
           df = otp.run(data)
           print(df)

        .. testoutput::

                                Time         ORIGIN_TIMESTAMP  PRICE  SIZE
           0 2022-03-02 00:00:00.000  2022-03-01 00:00:00.000    1.3   100
           1 2022-03-02 00:00:00.000  2022-03-01 00:00:00.002    1.4    50
           2 2022-03-02 00:00:00.000  2022-03-02 00:00:00.000    1.0   100
           3 2022-03-02 00:00:00.001  2022-03-02 00:00:00.001    1.1   101
           4 2022-03-02 00:00:00.002  2022-03-02 00:00:00.002    1.2   102
        """

        self.logger = otp.get_logger(__name__, self.__class__.__name__)

        if self._try_default_constructor(schema=schema, **kwargs):
            return

        schema = self._select_schema(schema, kwargs)

        if schema and (not schema_policy or schema_policy is utils.adaptive):
            schema_policy = self.POLICY_MANUAL

        if schema_policy is utils.adaptive:
            schema_policy = otp.config.default_schema_policy

        # for cases when we want to explicitly convert into string,
        # it might be symbol param or join_with_query parameter
        if isinstance(tick_type, OnetickParameter):
            tick_type = tick_type.parameter_expression

        if date:
            # TODO: write a warning in that case
            start, end = get_start_end_by_date(date)

        db, tick_type = self.__prepare_db_tick_type(db,
                                                    tick_type,
                                                    start,
                                                    end)

        self._p_db = db

        if not schema and schema_policy == self.POLICY_MANUAL_STRICT:
            raise ValueError(
                f"'{self.POLICY_MANUAL_STRICT}' schema policy was specified, but no schema has been provided"
            )

        self._p_strict = schema_policy in (self.POLICY_FAIL_STRICT,
                                           self.POLICY_TOLERANT_STRICT,
                                           self.POLICY_MANUAL_STRICT)

        # this is deprecated, but user may have set some complex types or values in schema,
        # let's infer basic onetick-py types from them
        for k, v in schema.items():
            schema[k] = ott.get_source_base_type(v)

        self._p_schema = self.__prepare_schema(db,  # tick type is embedded into the db
                                               start,
                                               schema_policy,
                                               guess_schema,
                                               schema)

        if symbols is not None:
            if symbol is utils.adaptive or symbol is None:
                symbol = symbols
            else:
                # TODO: test it
                raise ValueError('You have set the `symbol` and `symbols` parameters'
                                 'together, it is not allowed. Please, clarify parameters')

        if symbol_date is not None:
            if symbol is utils.adaptive or symbol is None:
                raise ValueError("Parameter 'symbol_date' can only be specified together with parameter 'symbols'")
            if isinstance(symbol, (str, list)):
                # this is a hack
                # onetick.query doesn't have an interface to set symbol_date for the EP node
                # so instead of setting symbols for the EP node,
                # we will turn symbol list into the first stage query, and symbol_date will be set for this query
                if isinstance(symbol, str):
                    symbol = [symbol]
                symbol = Ticks(SYMBOL_NAME=symbol)

        if isinstance(symbol, Symbols) and symbol._p_db is None:
            symbol = Symbols.duplicate(symbol, db=db)

        if identify_input_ts:
            if "SYMBOL_NAME" in schema:
                # TODO: think about how user could workaround it
                raise ValueError("Parameter 'identify_input_ts' is set,"
                                 " but field 'SYMBOL_NAME' is already in the schema")
            schema["SYMBOL_NAME"] = str
            self._p_schema["SYMBOL_NAME"] = str
            if "TICK_TYPE" in schema:
                raise ValueError("Parameter 'identify_input_ts' is set,"
                                 " but field 'TICK_TYPE' is already in the schema")
            schema["TICK_TYPE"] = str
            self._p_schema["TICK_TYPE"] = str

        # unobvious way to convert otp.Minute/Hour/... to number of seconds
        if type(back_to_first_tick).__name__ == '_DatePartCls':
            back_to_first_tick = int((ott.dt(0) + back_to_first_tick).timestamp())

        if isinstance(back_to_first_tick, _Operation):
            back_to_first_tick = otp.expr(back_to_first_tick)

        if back_to_first_tick != 0 and keep_first_tick_timestamp:
            schema[keep_first_tick_timestamp] = ott.nsectime
            self._p_schema[keep_first_tick_timestamp] = ott.nsectime

        if max_back_ticks_to_prepend < 1:
            raise ValueError(f'`max_back_ticks_to_prepend` must be at least 1 '
                             f'but {max_back_ticks_to_prepend} was passed')

        if where_clause_for_back_ticks is not None:
            # TODO: add otp.param here
            if not isinstance(where_clause_for_back_ticks, Raw):
                raise ValueError(f'Currently only otp.raw is supported for `where_clause_for_back_ticks` '
                                 f'but {type(where_clause_for_back_ticks)} was passed')
            if where_clause_for_back_ticks.dtype is not bool:
                raise ValueError(f'Only bool dtype for otp.raw in `where_clause_for_back_ticks` is supported '
                                 f'but {where_clause_for_back_ticks.dtype} was passed')
            where_clause_for_back_ticks = str(where_clause_for_back_ticks)

        self._schema = schema

        if isinstance(symbol, _QueryEvalWrapper):
            symbol_str = repr(symbol)
        else:
            symbol_str = symbol
        self.logger.info(
            otp.utils.json_dumps(dict(db=db, symbol=symbol_str, tick_type=tick_type, start=start, end=end))
        )

        if (
            isinstance(symbol, (Source, query, _QueryEvalWrapper, otq.GraphQuery))
            or hasattr(symbol, "__iter__")
            and not isinstance(symbol, (dict, str, OnetickParameter, _SymbolParamColumn))
        ):
            super().__init__(
                _start=start,
                _end=end,
                _base_ep_func=lambda: self._base_ep_for_cross_symbol(
                    db, symbol, tick_type,
                    identify_input_ts=identify_input_ts,
                    back_to_first_tick=back_to_first_tick,
                    keep_first_tick_timestamp=keep_first_tick_timestamp,
                    presort=presort, batch_size=batch_size, concurrency=concurrency,
                    max_back_ticks_to_prepend=max_back_ticks_to_prepend,
                    where_clause_for_back_ticks=where_clause_for_back_ticks,
                    symbol_date=symbol_date,
                ),
                schema=schema,
            )
        else:
            super().__init__(
                _symbols=symbol,
                _start=start,
                _end=end,
                _base_ep_func=lambda: self.base_ep(
                    db,
                    tick_type,
                    identify_input_ts=identify_input_ts,
                    back_to_first_tick=back_to_first_tick,
                    keep_first_tick_timestamp=keep_first_tick_timestamp,
                    max_back_ticks_to_prepend=max_back_ticks_to_prepend,
                    where_clause_for_back_ticks=where_clause_for_back_ticks,
                ),
                schema=schema,
            )

    @property
    def db(self):
        return self._p_db

    def _create_source(self, passthrough_ep, back_to_first_tick=0, keep_first_tick_timestamp=None):
        """Create graph that save original timestamp of first tick if needed"""
        if back_to_first_tick != 0 and keep_first_tick_timestamp:
            src = Source(otq.Passthrough(), schema=self._schema)
            src.sink(otq.AddField(field=keep_first_tick_timestamp, value='TIMESTAMP'))
            src.sink(passthrough_ep)
            return src
        return Source(passthrough_ep, schema=self._schema)

    def _table_schema(self, src):
        return src.table(**self._p_schema, strict=self._p_strict)

    def base_ep(
        self,
        db,
        tick_type,
        identify_input_ts,
        back_to_first_tick=0,
        keep_first_tick_timestamp=None,
        max_back_ticks_to_prepend=1,
        where_clause_for_back_ticks=None,
    ):
        str_db = convert_tick_type_to_str(tick_type, db)
        params = dict(
            go_back_to_first_tick=back_to_first_tick,
            max_back_ticks_to_prepend=max_back_ticks_to_prepend,
        )

        if where_clause_for_back_ticks is not None:
            params['where_clause_for_back_ticks'] = where_clause_for_back_ticks

        if isinstance(db, (list, _SymbolParamColumn)):
            src = self._create_source(otq.Passthrough(**params),
                                      back_to_first_tick=back_to_first_tick,
                                      keep_first_tick_timestamp=keep_first_tick_timestamp)

            if identify_input_ts or '+' in str_db or str_db.startswith('expr('):
                # PY-941: use MERGE only if we need to identify input or there are many databases,
                # otherwise use PASSTHROUGH, it seems to work faster in some cases
                src.sink(otq.Merge(identify_input_ts=identify_input_ts))
        else:
            if identify_input_ts:
                params["fields"] = "SYMBOL_NAME,TICK_TYPE"
                params["drop_fields"] = True

            src = self._create_source(otq.Passthrough(**params),
                                      back_to_first_tick=back_to_first_tick,
                                      keep_first_tick_timestamp=keep_first_tick_timestamp)
        src.tick_type(str_db)

        src = self._table_schema(src)
        return src

    def _cross_symbol_convert(self, symbol, symbol_date=None):
        tmp_otq = TmpOtq()

        if isinstance(symbol, _QueryEvalWrapper):
            symbol = symbol.to_eval_string(tmp_otq=tmp_otq, symbol_date=symbol_date)
        elif isinstance(symbol, query):
            if symbol_date is not None:
                raise ValueError("Parameter 'symbol_date' is not supported if symbols are set with otp.query object")
            symbol = symbol.to_eval_string()
        elif isinstance(symbol, (Source, otq.GraphQuery)):
            symbol = Source._convert_symbol_to_string(symbol, tmp_otq, symbol_date=symbol_date)

        return symbol, tmp_otq

    def _base_ep_for_cross_symbol(
        self, db, symbol, tick_type, identify_input_ts,
        back_to_first_tick=0, keep_first_tick_timestamp=None,
        presort=utils.adaptive, batch_size=None, concurrency=utils.default,
        max_back_ticks_to_prepend=1,
        where_clause_for_back_ticks=None,
        symbol_date=None,
    ):
        symbol, tmp_otq = self._cross_symbol_convert(symbol, symbol_date)

        self.logger.info(f'symbol={symbol}')

        tick_type = convert_tick_type_to_str(tick_type, db)

        kwargs = dict(
            go_back_to_first_tick=back_to_first_tick,
            max_back_ticks_to_prepend=max_back_ticks_to_prepend,
        )

        if where_clause_for_back_ticks is not None:
            kwargs['where_clause_for_back_ticks'] = where_clause_for_back_ticks

        src = self._create_source(otq.Passthrough(**kwargs),
                                  back_to_first_tick=back_to_first_tick,
                                  keep_first_tick_timestamp=keep_first_tick_timestamp)
        if presort is utils.adaptive:
            presort = True
        if presort:
            if batch_size is None:
                batch_size = otp.config.default_batch_size
            if concurrency is utils.default:
                concurrency = otp.configuration.default_presort_concurrency()
            if concurrency is None:
                # None means inherit concurrency from the query where this EP is used
                # otq.Presort does not support None
                concurrency = ''
            src.sink(
                otq.Presort(batch_size=batch_size, max_concurrency=concurrency).symbols(symbol).tick_type(tick_type)
            )
            src.sink(otq.Merge(identify_input_ts=identify_input_ts))
        else:
            src.sink(
                otq.Merge(identify_input_ts=identify_input_ts).symbols(symbol).tick_type(tick_type)
            )

        src._tmp_otq.merge(tmp_otq)

        src = self._table_schema(src)
        return src


Custom = DataSource  # for backward compatiblity, previously we had only Custom



from collections import defaultdict
from typing import List, Dict, Union, Optional, Tuple, TYPE_CHECKING
from copy import deepcopy

if TYPE_CHECKING:
    from onetick.py.core.source import Source   # hack for annotations

from onetick.py.core.column import _Column, _Operation
from onetick.py.core.column_operations.base import OnetickParameter
from onetick.py.core._source.tmp_otq import TmpOtq
from onetick.py import types as ott
from onetick.py.otq import otq

from ._base import (
    _Aggregation,
    _AggregationTSType,
    _AggregationTSSelection,
    _KeepTs,
    _FloatAggregation,
    _ExpectLargeInts,
    _ExpectDecimals,
    _MultiColumnAggregation,
    _AllColumnsAggregation,
    validate,
)


class First(_AggregationTSType, _ExpectLargeInts, _ExpectDecimals):
    NAME = "FIRST"
    EP = otq.First

    FIELDS_MAPPING = deepcopy(_AggregationTSType.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_ExpectLargeInts.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_ExpectDecimals.FIELDS_MAPPING)
    FIELDS_MAPPING['skip_tick_if'] = 'SKIP_TICK_IF'
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_ExpectLargeInts.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_ExpectDecimals.FIELDS_DEFAULT)
    FIELDS_DEFAULT['skip_tick_if'] = ''

    def __init__(self, *args, skip_tick_if=None, **kwargs):
        super().__init__(*args, **kwargs)

        if skip_tick_if is ott.nan:
            skip_tick_if = 'NAN'

        self.skip_tick_if = '' if skip_tick_if is None else skip_tick_if


class FirstTime(_AggregationTSType):
    NAME = "FIRST_TIME"
    EP = otq.FirstTime
    FIELDS_TO_SKIP = ["column_name"]

    def __init__(self, *args, **kwargs):
        super().__init__(column=_Column("TIMESTAMP"), *args, **kwargs)


class LastTime(_AggregationTSType):
    NAME = "LAST_TIME"
    EP = otq.LastTime
    FIELDS_TO_SKIP = ["column_name"]

    def __init__(self, *args, **kwargs):
        super().__init__(column=_Column("TIMESTAMP"), *args, **kwargs)


class Last(First):
    NAME = "LAST"
    EP = otq.Last

    FIELDS_MAPPING = deepcopy(First.FIELDS_MAPPING)
    FIELDS_MAPPING['skip_tick_if'] = 'FWD_FILL_IF'
    FIELDS_DEFAULT = deepcopy(First.FIELDS_DEFAULT)
    FIELDS_DEFAULT['skip_tick_if'] = ''


class Count(_Aggregation):
    NAME = "NUM_TICKS"
    EP = otq.NumTicks
    FIELDS_TO_SKIP = ["column_name"]
    output_field_type = int

    def __init__(self, *args, **kwargs):
        super().__init__(column=_Column("TIMESTAMP"), *args, **kwargs)

    def apply(self, src, name: str = 'VALUE', *args, **kwargs):
        return super().apply(src=src, name=name, *args, **kwargs)


class Vwap(_Aggregation):

    NAME = "VWAP"
    EP = otq.Vwap

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['price_column'] = 'PRICE_FIELD_NAME'
    FIELDS_MAPPING['size_column'] = 'SIZE_FIELD_NAME'
    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['price_column'] = 'PRICE'
    FIELDS_DEFAULT['size_column'] = 'SIZE'

    FIELDS_TO_SKIP: List = ['column_name']

    output_field_type = float
    require_type = (int, float, ott.nsectime, ott.decimal)

    def __init__(self,
                 price_column: str,
                 size_column: str,
                 *args, **kwargs):
        super().__init__(column=_Column('TIMESTAMP'), *args, **kwargs)  # type: ignore
        self.price_column = str(price_column)
        self.size_column = str(size_column)

    def validate_input_columns(self, src: 'Source'):
        for column in [self.price_column, self.size_column]:
            if column not in src.schema:
                raise TypeError(
                    f"Aggregation `{self.NAME}` uses column `{column}` as input, which doesn't exist")
            if not issubclass(src.schema[column], self.require_type):
                raise TypeError(f"Aggregation `{self.NAME}` require {self.require_type} types, "
                                f"got {src.schema[column]}")

    def apply(self, src: 'Source', name: str = 'VWAP', *args, **kwargs) -> 'Source':
        return super().apply(src=src, name=name, *args, **kwargs)


class Correlation(_Aggregation):

    NAME = 'CORRELATION'
    EP = otq.Correlation

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['column_name_1'] = 'INPUT_FIELD1_NAME'
    FIELDS_MAPPING['column_name_2'] = 'INPUT_FIELD2_NAME'

    FIELDS_TO_SKIP: List = ['column_name']

    output_field_type = float
    require_type = (int, float)

    def __init__(self,
                 column_name_1: str,
                 column_name_2: str,
                 *args, **kwargs):
        super().__init__(column=_Column('TIMESTAMP'), *args, **kwargs)  # type: ignore
        self.column_name_1 = str(column_name_1)
        self.column_name_2 = str(column_name_2)

    def validate_input_columns(self, src: 'Source'):
        for column in [self.column_name_1, self.column_name_2]:
            if column not in src.schema:
                raise TypeError(
                    f"Aggregation `{self.NAME}` uses column `{column}` as input, which doesn't exist")
            if src.schema[column] not in self.require_type:
                raise TypeError(f"Aggregation `{self.NAME}` require {self.require_type} types, "
                                f"got {src.schema[column]}")

    def apply(self, src: 'Source', name: str = 'CORRELATION', *args, **kwargs) -> 'Source':
        return super().apply(src=src, name=name, *args, **kwargs)


class FirstTick(_AggregationTSType, _KeepTs, _AllColumnsAggregation):
    EP = otq.FirstTick
    NAME = 'FIRST_TICK'
    DEFAULT_OUTPUT_NAME = 'FIRST_TICK'

    FIELDS_MAPPING = deepcopy(_AggregationTSType.FIELDS_MAPPING)
    FIELDS_MAPPING['n'] = 'NUM_TICKS'
    FIELDS_MAPPING['default_tick'] = 'DEFAULT_TICK'
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT['n'] = 1
    FIELDS_DEFAULT['default_tick'] = ''

    FIELDS_TO_SKIP = ['column_name', 'output_field_name', 'all_fields']

    _validations_to_skip = ['running_all_fields']

    def __init__(self, n: int = 1, *args, default_tick=None, **kwargs):
        kwargs['all_fields'] = True
        super().__init__(_Column("TIMESTAMP"), *args, **kwargs)
        self.n = n
        self._default_tick = default_tick
        self.default_tick = ''

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)
        if not self._default_tick:
            self.default_tick = ''
            return

        fields = {}
        for field, value in self._default_tick.items():
            if field not in src.schema:
                raise ValueError(f"Field '{field}' is not in schema")
            dtype = src.schema[field]
            if value is not None:
                dtype = ott.get_object_type(value)
                if dtype is not src.schema[field]:
                    raise ValueError(f"Incompatible types for field '{field}': {src.schema[field]} --> {dtype}")
                value = ott.value2str(value)
            dtype_str = ott.type2str(dtype)
            fields[field] = (dtype_str, value)
        self.default_tick = ','.join(
            f'{field} {dtype} ({value})' if value is not None else f'{field} {dtype}'
            for field, (dtype, value) in fields.items()
        )


class LastTick(FirstTick):
    EP = otq.LastTick
    NAME = 'LAST_TICK'
    DEFAULT_OUTPUT_NAME = 'LAST_TICK'


class Distinct(_AggregationTSSelection):
    EP = otq.Distinct
    NAME = 'DISTINCT'

    FIELDS_MAPPING = deepcopy(_AggregationTSSelection.FIELDS_MAPPING)
    FIELDS_MAPPING['column_name'] = 'KEYS'
    FIELDS_MAPPING['key_attrs_only'] = 'KEY_ATTRS_ONLY'
    FIELDS_DEFAULT = deepcopy(_AggregationTSSelection.FIELDS_DEFAULT)
    FIELDS_DEFAULT['key_attrs_only'] = True

    FIELDS_TO_SKIP = ['end_condition_per_group', 'group_by', 'output_field_name', 'all_fields']

    def __init__(self,
                 keys: Union[str, List[str], _Column, List[_Column]],
                 key_attrs_only: bool = True,
                 *args, **kwargs):
        keys = keys if isinstance(keys, list) else [keys]   # type: ignore
        super().__init__(column=keys, *args, **kwargs)  # type: ignore
        self.key_attrs_only = key_attrs_only

    @validate
    def _get_common_schema(self, src: 'Source', *args, **kwargs) -> Dict:
        if self.key_attrs_only:
            return super()._get_common_schema(src=src, *args, **kwargs)
        return src.schema.copy()

    @staticmethod
    def validate_output_name(*args, **kwargs):
        # Distinct aggregation doesn't have output fields
        pass

    def validate_input_columns(self, src: 'Source'):
        for col in str(self.column_name).split(','):
            if col.strip() not in src.schema:
                raise TypeError(f"Aggregation `{self.NAME}` uses column `{col.strip()}` as input, which doesn't exist")

    def apply(self, src: 'Source', *args, **kwargs) -> 'Source':
        res = src.copy()
        res.sink(self.to_ep(name=None))
        schema = self._get_common_schema(src, None)
        for col in str(self.column_name).split(','):
            schema[col.strip()] = src.schema[col.strip()]
        res.schema.set(**schema)
        return res


class Sum(_FloatAggregation, _ExpectDecimals):
    NAME = "SUM"
    EP = otq.Sum

    FIELDS_MAPPING = deepcopy(_FloatAggregation.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_ExpectDecimals.FIELDS_MAPPING)
    FIELDS_DEFAULT = deepcopy(_FloatAggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_ExpectDecimals.FIELDS_DEFAULT)


class Average(_FloatAggregation):
    NAME = "AVERAGE"
    EP = otq.Average
    output_field_type = float


class StdDev(_Aggregation):     # Stddev does not support inf, so no need to use _FloatAggregation
    NAME = "STDDEV"
    EP = otq.Stddev
    require_type = (int, float, ott.decimal)
    output_field_type = float
    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['biased'] = 'BIASED'

    def __init__(self, *args, biased: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.biased = biased


class TimeWeightedAvg(_AggregationTSType, _FloatAggregation):
    NAME = "TW_AVERAGE"
    EP = otq.TwAverage
    output_field_type = float
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT['time_series_type'] = 'state_ts'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('time_series_type', 'state_ts')
        super().__init__(*args, **kwargs)


class Median(_FloatAggregation, _ExpectDecimals):
    NAME = "MEDIAN"
    EP = otq.Median
    output_field_type = float

    FIELDS_MAPPING = deepcopy(_FloatAggregation.FIELDS_MAPPING)
    FIELDS_MAPPING.update(_ExpectDecimals.FIELDS_MAPPING)
    FIELDS_DEFAULT = deepcopy(_FloatAggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update(_ExpectDecimals.FIELDS_DEFAULT)


class OptionPrice(_Aggregation):

    NAME = 'OPTION_PRICE'
    EP = otq.OptionPrice

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['volatility'] = 'VOLATILITY'
    FIELDS_MAPPING['interest_rate'] = 'INTEREST_RATE'
    FIELDS_MAPPING['compute_model'] = 'COMPUTE_MODEL'
    FIELDS_MAPPING['number_of_steps'] = 'NUMBER_OF_STEPS'
    FIELDS_MAPPING['compute_delta'] = 'COMPUTE_DELTA'
    FIELDS_MAPPING['compute_gamma'] = 'COMPUTE_GAMMA'
    FIELDS_MAPPING['compute_theta'] = 'COMPUTE_THETA'
    FIELDS_MAPPING['compute_vega'] = 'COMPUTE_VEGA'
    FIELDS_MAPPING['compute_rho'] = 'COMPUTE_RHO'
    FIELDS_MAPPING['volatility_field_name'] = 'VOLATILITY_FIELD_NAME'
    FIELDS_MAPPING['interest_rate_field_name'] = 'INTEREST_RATE_FIELD_NAME'
    FIELDS_MAPPING['option_type_field_name'] = 'OPTION_TYPE_FIELD_NAME'
    FIELDS_MAPPING['strike_price_field_name'] = 'STRIKE_PRICE_FIELD_NAME'
    FIELDS_MAPPING['days_in_year'] = 'DAYS_IN_YEAR'
    FIELDS_MAPPING['days_till_expiration_field_name'] = 'DAYS_TILL_EXPIRATION_FIELD_NAME'
    FIELDS_MAPPING['expiration_date_field_name'] = 'EXPIRATION_DATE_FIELD_NAME'

    FIELDS_TO_SKIP: List = ['column_name', 'all_fields', 'end_condition_per_group', 'group_by', 'output_field_name']

    output_field_type = float
    require_type = (float,)

    def __init__(self,  # NOSONAR
                 volatility: Optional[float] = None,
                 interest_rate: Optional[float] = None,
                 compute_model: str = 'BS',
                 number_of_steps: Optional[int] = None,
                 compute_delta: bool = False,
                 compute_gamma: bool = False,
                 compute_theta: bool = False,
                 compute_vega: bool = False,
                 compute_rho: bool = False,
                 volatility_field_name: str = '',
                 interest_rate_field_name: str = '',
                 option_type_field_name: str = '',
                 strike_price_field_name: str = '',
                 days_in_year: int = 365,
                 days_till_expiration_field_name: str = '',
                 expiration_date_field_name: str = '',
                 *args, **kwargs):
        super().__init__(column=_Column('PRICE'), *args, **kwargs)  # type: ignore
        self.volatility = volatility
        self.interest_rate = interest_rate
        self.compute_model = compute_model
        self.number_of_steps = number_of_steps
        self.compute_delta = compute_delta
        self.compute_gamma = compute_gamma
        self.compute_theta = compute_theta
        self.compute_vega = compute_vega
        self.compute_rho = compute_rho
        self.volatility_field_name = volatility_field_name
        self.interest_rate_field_name = interest_rate_field_name
        self.option_type_field_name = option_type_field_name
        self.strike_price_field_name = strike_price_field_name
        self.days_in_year = days_in_year
        self.days_till_expiration_field_name = days_till_expiration_field_name
        self.expiration_date_field_name = expiration_date_field_name

    def apply(self, src: 'Source', name: str = 'VALUE', *args, **kwargs) -> 'Source':
        return super().apply(src=src, name=name, *args, **kwargs)

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> Dict:
        output_schema = super()._get_output_schema(src, name=name)
        compute = {
            k.upper(): float for k in ['delta', 'gamma', 'theta', 'vega', 'rho'] if getattr(self, f'compute_{k}')
        }
        output_schema.update(compute)
        return output_schema


class Ranking(_KeepTs):
    NAME = 'RANKING'
    EP = otq.Ranking
    FIELDS_TO_SKIP = ['column_name', 'output_field_name']
    FIELDS_MAPPING = deepcopy(_KeepTs.FIELDS_MAPPING)
    FIELDS_MAPPING.update({
        'rank_by': 'RANK_BY',
        'show_rank_as': 'SHOW_RANK_AS',
        'include_tick_in_percentage': 'INCLUDE_TICK_IN_PERCENTAGE',
    })
    FIELDS_DEFAULT = deepcopy(_KeepTs.FIELDS_DEFAULT)
    FIELDS_DEFAULT.update({
        'show_rank_as': 'ORDER',
        'include_tick_in_percentage': False,
    })
    _ALLOWED_RANK_BY = {'asc', 'desc'}
    _ALLOWED_SHOW_RANK_AS = {'order', 'percentile_standard', 'percent_le_values', 'percent_lt_values'}

    def __init__(self, rank_by, *args, show_rank_as='ORDER', include_tick=False, **kwargs):
        super().__init__(column=_Column('TIMESTAMP'), *args, **kwargs)

        if isinstance(rank_by, str):
            rank_by = [rank_by]
        if not isinstance(rank_by, dict):
            rank_by = {column: 'desc' for column in rank_by}
        for k in rank_by:
            rank_by[k] = rank_by[k].lower()
        rank_by_values = set(rank_by.values())
        diff = rank_by_values.difference(self._ALLOWED_RANK_BY)
        if diff:
            raise ValueError(f"Only {self._ALLOWED_RANK_BY} values of parameter 'rank_by' are allowed, got {diff}.")
        self._rank_by = rank_by

        show_rank_as = show_rank_as.lower()
        if show_rank_as not in self._ALLOWED_SHOW_RANK_AS:
            raise ValueError(f"Only {self._ALLOWED_SHOW_RANK_AS} values of parameter 'show_rank_as' are allowed.")
        self.show_rank_as = show_rank_as
        self.include_tick_in_percentage = include_tick

    @property
    def rank_by(self):
        return ','.join(f'{k} {w}' for k, w in self._rank_by.items())

    def apply(self, src):
        return super().apply(src=src, name='RANKING')

    def validate_input_columns(self, src: 'Source'):
        for col in self._rank_by:
            if col not in src.schema:
                raise TypeError(f"Ranking aggregation uses column '{col}' as input, which doesn't exist")

    @staticmethod
    def validate_output_name(schema, name):
        if 'RANKING' in schema:
            raise ValueError("Output field of ranking aggregation will be named 'RANKING',"
                             " but this field exists in the input schema.")
        _KeepTs.validate_output_name(schema, name)

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None):
        return {
            'RANKING': int if self.show_rank_as == 'order' else float
        }


class Variance(_Aggregation):
    NAME = 'VARIANCE'
    EP = otq.Variance
    output_field_type = float
    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['biased'] = 'BIASED'
    require_type = (int, float)

    def __init__(self,
                 column: Union[str, _Column, _Operation],
                 biased: bool,
                 *args, **kwargs):
        super().__init__(column, *args, **kwargs)
        self.biased = biased

    def apply(self, src: 'Source', name: str = 'VARIANCE', *args, **kwargs) -> 'Source':
        return super().apply(src=src, name=name, *args, **kwargs)


class Percentile(_Aggregation):
    NAME = 'PERCENTILE'
    EP = otq.Percentile

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['number_of_quantiles'] = 'NUMBER_OF_QUANTILES'
    FIELDS_MAPPING['column_name'] = 'INPUT_FIELD_NAMES'
    FIELDS_MAPPING['output_field_names'] = 'OUTPUT_FIELD_NAMES'

    FIELDS_TO_SKIP = ['output_field_name', 'all_fields']

    def __init__(
        self,
        input_field_names: List[Union[Union[str, _Column], Tuple[Union[str, _Column], str]]],
        output_field_names: Optional[List[str]] = None,
        number_of_quantiles: int = 2,
        *args, **kwargs,
    ):
        self.number_of_quantiles = number_of_quantiles

        if not input_field_names:
            raise ValueError('`input_field_names` should be passed')

        if not isinstance(input_field_names, list):
            raise ValueError('`input_field_names` should be the `list` of columns')

        if output_field_names is not None and not isinstance(output_field_names, list):
            raise ValueError('`output_field_names` should be the `list` of columns')

        if output_field_names and len(input_field_names) != len(output_field_names):
            raise ValueError('`output_field_names` should have the same number of elements as `input_field_names`')

        columns = ', '.join(
            str(field) if isinstance(field, (str, _Column)) else f'{str(field[0])} {field[1].upper()}'
            for field in input_field_names
        )

        self._columns = [
            str(field) if not isinstance(field, tuple) else str(field[0])
            for field in input_field_names
        ]

        self.output_field_names = '' if output_field_names is None else ', '.join(output_field_names)

        super().__init__(columns, *args, **kwargs)

    def validate_input_columns(self, src: 'Source'):
        for col in self._columns:
            if col.strip() not in src.schema:
                raise TypeError(f"Aggregation `{self.NAME}` uses column `{col.strip()}` as input, which doesn't exist")

    def apply(self, src: 'Source', *args, **kwargs) -> 'Source':
        res = src.copy()
        res.sink(self.to_ep(name=None))
        schema = self._get_common_schema(src, None)
        for col in self._columns:
            schema[col.strip()] = src.schema[col.strip()]

        schema['QUANTILE'] = int
        res.schema.set(**schema)
        return res


class FindValueForPercentile(_FloatAggregation):
    NAME = 'FIND_VALUE_FOR_PERCENTILE'
    EP = None

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['percentile'] = 'PERCENTILE'
    FIELDS_MAPPING['show_percentile_as'] = 'SHOW_PERCENTILE_AS'

    FIELDS_TO_SKIP = ['all_fields']

    output_field_type = float

    def __init__(
        self, column,
        percentile: int,
        show_percentile_as: str = '',
        *args, **kwargs
    ):
        try:
            self.EP = otq.FindValueForPercentile
        except AttributeError as exc:
            raise RuntimeError("This OneTick build doesn't support FindValueForPercentile aggregation") from exc

        if not isinstance(percentile, int) or percentile < 0 or percentile > 100:
            raise ValueError("Parameter 'percentile' must be a number between 0 and 100.")
        if show_percentile_as and show_percentile_as not in {'interpolated_value', 'first_value_with_ge_percentile'}:
            raise ValueError(f"Unsupported value for parameter 'show_percentile_as': {show_percentile_as}")
        self.percentile = percentile
        self.show_percentile_as = show_percentile_as.upper()

        FindValueForPercentile.FIELDS_MAPPING['show_percentile_as'] = \
            'SHOW_PERCENTILE_AS' if 'show_percentile_as' in self.EP.Parameters.__dict__ else 'COMPUTE_VALUE_AS'

        super().__init__(column, *args, **kwargs)


class ExpWAverage(_FloatAggregation, _AggregationTSType):
    NAME = 'EXP_W_AVERAGE'
    EP = otq.ExpWAverage

    FIELDS_MAPPING = deepcopy(_AggregationTSType.FIELDS_MAPPING)
    FIELDS_MAPPING['decay'] = 'DECAY'
    FIELDS_MAPPING['decay_value_type'] = 'DECAY_VALUE_TYPE'
    FIELDS_DEFAULT = deepcopy(_AggregationTSType.FIELDS_DEFAULT)
    FIELDS_DEFAULT['time_series_type'] = 'state_ts'
    FIELDS_DEFAULT['decay_value_type'] = 'lambda'

    output_field_type = float

    def __init__(self, column, decay, decay_value_type='lambda', time_series_type='state_ts', *args, **kwargs):
        super().__init__(column, time_series_type=time_series_type, *args, **kwargs)

        if decay_value_type not in ['lambda', 'half_life_index']:
            raise ValueError(f"Parameter 'decay_value_type' has incorrect value: {decay_value_type}, "
                             f"should be one of next: 'lambda', 'half_life_index'")

        self.decay = decay
        self.decay_value_type = decay_value_type.upper()


class ExpTwAverage(_FloatAggregation):
    NAME = 'EXP_TW_AVERAGE'
    EP = otq.ExpTwAverage

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['decay'] = 'DECAY'
    FIELDS_MAPPING['decay_value_type'] = 'DECAY_VALUE_TYPE'
    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['decay_value_type'] = 'half_life_index'

    output_field_type = float

    def __init__(self, column, decay, decay_value_type='half_life_index', *args, **kwargs):
        if decay_value_type not in ['lambda', 'half_life_index']:
            raise ValueError(f"Parameter 'decay_value_type' has incorrect value: {decay_value_type}, "
                             f"should be one of next: 'lambda', 'half_life_index'")

        self.decay = decay
        self.decay_value_type = decay_value_type.upper()

        super().__init__(column, *args, **kwargs)


class StandardizedMoment(_Aggregation):
    NAME = 'STANDARDIZED_MOMENT'
    EP = None

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['degree'] = 'DEGREE'
    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['degree'] = 3

    output_field_type = float

    def __init__(self, *args, degree=3, **kwargs):
        try:
            self.EP = otq.StandardizedMoment
        except AttributeError as exc:
            raise RuntimeError("Used onetick installation not support onetick.query.StandardizedMoment") from exc

        self.degree = degree

        super().__init__(*args, **kwargs)


class PortfolioPrice(_Aggregation, _MultiColumnAggregation):
    NAME = 'PORTFOLIO_PRICE'
    EP = otq.PortfolioPrice
    DEFAULT_OUTPUT_NAME = 'VALUE'

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['column_name'] = 'INPUT_FIELD_NAME'
    FIELDS_MAPPING['weight_field_name'] = 'WEIGHT_FIELD_NAME'
    FIELDS_MAPPING['side'] = 'SIDE'
    FIELDS_MAPPING['weight_type'] = 'WEIGHT_TYPE'

    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['column_name'] = ''
    FIELDS_DEFAULT['weight_field_name'] = ''
    FIELDS_DEFAULT['side'] = 'BOTH'
    FIELDS_DEFAULT['weight_type'] = 'ABSOLUTE'

    FIELDS_TO_SKIP = ['output_field_name']

    output_field_type = float

    def __init__(
        self, column='', weight_field_name='', side='both', weight_type='absolute', symbols=None, *args, **kwargs,
    ):
        if side not in ['long', 'short', 'both']:
            raise ValueError(f"Parameter `side` has incorrect value: {side}, "
                             f"should be one of next: 'long', 'short', 'both")

        if weight_type not in ['absolute', 'relative']:
            raise ValueError(f"Parameter `weight_type` has incorrect value: {weight_type}, "
                             f"should be one of next: 'absolute', 'relative'")

        if isinstance(weight_field_name, _Column):
            weight_field_name = str(weight_field_name)

        self.weight_field_name = weight_field_name
        self.side = side.upper()
        self.weight_type = weight_type.upper()
        self.symbols = symbols

        super().__init__(column=column, *args, **kwargs)

    def apply(self, src, name='VALUE', inplace=False):
        if inplace:
            res = src
            src = src.copy()
        else:
            res = src.copy()

        out_name = name or self.column_name
        schema = self._get_common_schema(src, out_name)
        # it's important to validate input schema before sinking
        self._modify_source(res)

        ep = self.to_ep(name=str(out_name))

        if self.symbols:
            from onetick.py.core.source import Source
            from onetick.py.sources import query as otp_query
            from onetick.py.core.eval_query import _QueryEvalWrapper

            tmp_otq = TmpOtq()

            if isinstance(self.symbols, (otp_query, _QueryEvalWrapper)):
                self.symbols = self.symbols.to_eval_string(tmp_otq=tmp_otq)
            elif isinstance(self.symbols, (Source, otq.GraphQuery)):
                self.symbols = Source._convert_symbol_to_string(self.symbols, tmp_otq=tmp_otq)

            ep = ep.symbols(self.symbols)

            new_res = Source(schema=schema)

            new_res._copy_state_vars_from(res)
            new_res._clean_sources_dates()
            new_res._merge_tmp_otq(res)

            if isinstance(self.symbols, Source):
                new_res._merge_tmp_otq(self.symbols)

            new_res._tmp_otq.merge(tmp_otq)

            eps = defaultdict()
            new_res.source(res.node().copy_graph(eps))
            new_res.node().add_rules(res.node().copy_rules())
            new_res._set_sources_dates(res, copy_symbols=not bool(self.symbols))

            res = new_res

            if inplace:
                src = res

        res.sink(ep)

        schema.update(self._get_output_schema(src, str(out_name)))
        res.schema.set(**schema)
        if not self.all_fields:
            # in this case we propagate only resulting fields, that stored in res.schema (flexible schema case)
            res._add_table(strict=True)
        else:
            # adding table to convert types in schema, e.g. float to int
            res._add_table(strict=False)

        return res

    def validate_input_columns(self, src: 'Source'):
        if self.weight_field_name and self.weight_field_name not in src.schema:
            raise TypeError(
                f"Aggregation `{self.NAME}` uses column `{self.weight_field_name}` as input, which doesn't exist",
            )

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None):
        if not name:
            name = self.DEFAULT_OUTPUT_NAME

        return {
            name: float,
            'NUM_SYMBOLS': int,
        }


class MultiPortfolioPrice(PortfolioPrice):
    NAME = 'MULTI_PORTFOLIO_PRICE'
    EP = otq.MultiPortfolioPrice
    DEFAULT_OUTPUT_NAME = 'VALUE'

    FIELDS_MAPPING = deepcopy(PortfolioPrice.FIELDS_MAPPING)
    FIELDS_MAPPING['weight_multiplier_field_name'] = 'WEIGHT_MULTIPLIER_FIELD_NAME'
    FIELDS_MAPPING['portfolios_query'] = 'PORTFOLIOS_QUERY'
    FIELDS_MAPPING['portfolios_query_params'] = 'PORTFOLIOS_QUERY_PARAMS'
    FIELDS_MAPPING['portfolio_value_field_name'] = 'PORTFOLIO_VALUE_FIELD_NAME'

    FIELDS_DEFAULT = deepcopy(PortfolioPrice.FIELDS_DEFAULT)
    FIELDS_DEFAULT['weight_multiplier_field_name'] = ''
    FIELDS_DEFAULT['portfolios_query'] = ''
    FIELDS_DEFAULT['portfolios_query_params'] = ''
    FIELDS_DEFAULT['portfolio_value_field_name'] = ''

    FIELDS_TO_SKIP = ['output_field_name']

    output_field_type = float

    def __init__(
        self, columns='', weight_multiplier_field_name='', portfolios_query='', portfolios_query_params='',
        portfolio_value_field_name='', *args, **kwargs,
    ):
        if not portfolios_query:
            raise ValueError('Required parameter `portfolio_query` not set')

        self.weight_multiplier_field_name = weight_multiplier_field_name

        from onetick.py.core.source import Source
        if isinstance(portfolios_query, Source):
            self.portfolios_query = portfolios_query.to_otq()
        else:
            self.portfolios_query = portfolios_query

        if isinstance(portfolios_query_params, dict):
            portfolios_query_params = ','.join([f'{k}={ott.value2str(v)}' for k, v in portfolios_query_params.items()])

        self.portfolios_query_params = portfolios_query_params

        if not portfolio_value_field_name:
            portfolio_value_field_name = self.DEFAULT_OUTPUT_NAME

        if isinstance(portfolio_value_field_name, str):
            portfolio_value_field_name_str = portfolio_value_field_name
            portfolio_value_field_name_list = portfolio_value_field_name.split(',')
        else:
            portfolio_value_field_name_str = ','.join(map(str, portfolio_value_field_name))
            portfolio_value_field_name_list = list(map(str, portfolio_value_field_name))

        self._portfolio_value_field_name = portfolio_value_field_name_list
        self.portfolio_value_field_name = portfolio_value_field_name_str

        len_columns = max(len(columns) if isinstance(columns, list) else len(columns.split(',')), 1)
        if len_columns != len(portfolio_value_field_name_list):
            raise RuntimeError('The number of the field names must match the number of the field names '
                               'listed in the `columns` parameter')

        super().__init__(column=columns, *args, **kwargs)

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)

        if self.weight_multiplier_field_name and self.weight_multiplier_field_name not in src.schema:
            raise TypeError(
                f"Aggregation `{self.NAME}` uses column `{self.weight_multiplier_field_name}` as input, "
                f"which doesn't exist",
            )

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None):
        if not name:
            return super()._get_output_schema(src, name)

        return {
            **{column_name: float for column_name in self._portfolio_value_field_name},
            'NUM_SYMBOLS': int,
            'PORTFOLIO_NAME': str,
        }


class Return(_Aggregation):
    NAME = 'RETURN'
    EP = otq.Return
    DEFAULT_OUTPUT_NAME = 'RETURN'

    output_field_type = float


class ImpliedVol(_Aggregation):
    NAME = 'IMPLIED_VOL'
    EP = otq.ImpliedVol

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['interest_rate'] = 'INTEREST_RATE'
    FIELDS_MAPPING['price_field_name'] = 'PRICE_FIELD_NAME'
    FIELDS_MAPPING['option_price_field_name'] = 'OPTION_PRICE_FIELD_NAME'
    FIELDS_MAPPING['method'] = 'METHOD'
    FIELDS_MAPPING['precision'] = 'PRECISION'
    FIELDS_MAPPING['value_for_non_converge'] = 'VALUE_FOR_NON_CONVERGE'
    FIELDS_MAPPING['interest_rate_field_name'] = 'INTEREST_RATE_FIELD_NAME'
    FIELDS_MAPPING['option_type_field_name'] = 'OPTION_TYPE_FIELD_NAME'
    FIELDS_MAPPING['strike_price_field_name'] = 'STRIKE_PRICE_FIELD_NAME'
    FIELDS_MAPPING['days_in_year'] = 'DAYS_IN_YEAR'
    FIELDS_MAPPING['days_till_expiration_field_name'] = 'DAYS_TILL_EXPIRATION_FIELD_NAME'
    FIELDS_MAPPING['expiration_date_field_name'] = 'EXPIRATION_DATE_FIELD_NAME'

    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['interest_rate'] = None
    FIELDS_DEFAULT['price_field_name'] = 'PRICE'
    FIELDS_DEFAULT['option_price_field_name'] = 'OPTION_PRICE'
    FIELDS_DEFAULT['method'] = 'NEWTON'
    FIELDS_DEFAULT['precision'] = 0.00001
    FIELDS_DEFAULT['value_for_non_converge'] = 'NAN_VAL'
    FIELDS_DEFAULT['interest_rate_field_name'] = ''
    FIELDS_DEFAULT['option_type_field_name'] = ''
    FIELDS_DEFAULT['strike_price_field_name'] = ''
    FIELDS_DEFAULT['days_in_year'] = 365
    FIELDS_DEFAULT['days_till_expiration_field_name'] = ''
    FIELDS_DEFAULT['expiration_date_field_name'] = ''

    FIELDS_TO_SKIP = ['column_name', 'output_field_name']

    output_field_type = float

    def __init__(
        self,
        interest_rate: Optional[Union[float, int, _Column, str]] = None,
        price_field: Union[_Column, str] = 'PRICE',
        option_price_field: Union[_Column, str] = 'OPTION_PRICE',
        method: str = 'newton',
        precision: float = 0.00001,
        value_for_non_converge: str = 'nan_val',
        option_type_field: Union[_Column, str] = '',
        strike_price_field: Union[_Column, str] = '',
        days_in_year: int = 365,
        days_till_expiration_field: Union[_Column, str] = '',
        expiration_date_field: Union[_Column, str] = '',
        *args, **kwargs,
    ):
        self.interest_rate_field_name = ''
        self.interest_rate = None

        if interest_rate is not None:
            if isinstance(interest_rate, (str, _Column)):
                self.interest_rate_field_name = str(interest_rate)
            elif isinstance(interest_rate, (int, float)):
                self.interest_rate = interest_rate
            else:
                raise ValueError(f'Unsupported value passed as `interest_rate`: {type(interest_rate)}')

        if value_for_non_converge not in {'nan_val', 'closest_found_val'}:
            raise ValueError(f'Unsupported value for `value_for_non_converge` was passed: {value_for_non_converge}')

        if method not in {'newton', 'newton_with_fallback', 'bisections'}:
            raise ValueError(f'Unsupported value for `method` was passed: {method}')

        if isinstance(price_field, _Column):
            price_field = str(price_field)
        if isinstance(option_price_field, _Column):
            option_price_field = str(option_price_field)

        if isinstance(option_type_field, _Column):
            option_type_field = str(option_type_field)
        if isinstance(strike_price_field, _Column):
            strike_price_field = str(strike_price_field)
        if isinstance(days_till_expiration_field, _Column):
            days_till_expiration_field = str(days_till_expiration_field)
        if isinstance(expiration_date_field, _Column):
            expiration_date_field = str(expiration_date_field)

        self.price_field_name = price_field
        self.option_price_field_name = option_price_field
        self.method = method.upper()
        self.precision = precision
        self.value_for_non_converge = value_for_non_converge.upper()
        self.option_type_field_name = option_type_field
        self.strike_price_field_name = strike_price_field
        self.days_in_year = days_in_year
        self.days_till_expiration_field_name = days_till_expiration_field
        self.expiration_date_field_name = expiration_date_field

        super().__init__(_Column(price_field), *args, **kwargs)

    def validate_input_columns(self, src: 'Source'):
        columns_to_check: Dict[str, Tuple[Union[str, _Column], Union[Tuple, type]]] = {
            'price_field': (self.price_field_name, (int, float)),
            'option_price_field': (self.option_price_field_name, (int, float)),
            'interest_rate': (self.interest_rate_field_name, (int, float)),
            'option_type_field': (self.option_type_field_name, str),
            'strike_price_field': (self.strike_price_field_name, (int, float)),
            'days_till_expiration_field': (self.days_till_expiration_field_name, int),
            'expiration_date_field': (self.expiration_date_field_name, int),
        }

        for parameter, (column, required_type) in columns_to_check.items():
            if not column:
                # probably parameter will be set as symbol parameter, we couldn't check this right now
                continue

            if column not in src.schema:
                raise TypeError(
                    f"Aggregation `{self.NAME}` uses column `{column}` from parameter `{parameter}` as input, "
                    f"which doesn't exist."
                )

            if not issubclass(src.schema[column], required_type):
                raise TypeError(
                    f"Aggregation `{self.NAME}` require column `{column}` from parameter `{parameter}` "
                    f"to be {required_type}, got {src.schema[column]}"
                )

    def apply(self, src: 'Source', name: str = 'VALUE', *args, **kwargs) -> 'Source':
        return super().apply(src=src, name=name, *args, **kwargs)


class LinearRegression(_Aggregation, _MultiColumnAggregation):
    NAME = 'LINEAR_REGRESSION'
    EP = None

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['dependent_variable_field_name'] = 'DEPENDENT_VARIABLE_FIELD_NAME'
    FIELDS_MAPPING['independent_variable_field_name'] = 'INDEPENDENT_VARIABLE_FIELD_NAME'

    FIELDS_TO_SKIP = ['column_name', 'output_field_name']

    def __init__(
        self,
        dependent_variable_field_name: Union[_Column, str, OnetickParameter],
        independent_variable_field_name: Union[_Column, str, OnetickParameter],
        *args, **kwargs,
    ):
        try:
            self.EP = otq.LinearRegression
        except AttributeError as exc:
            raise RuntimeError("Used onetick installation not support onetick.query.LinearRegression") from exc

        if isinstance(dependent_variable_field_name, _Column):
            dependent_variable_field_name = str(dependent_variable_field_name)
        if isinstance(independent_variable_field_name, _Column):
            independent_variable_field_name = str(independent_variable_field_name)

        self.dependent_variable_field_name = dependent_variable_field_name
        self.independent_variable_field_name = independent_variable_field_name

        super().__init__(_Column('TIMESTAMP'), *args, **kwargs)

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)

        for column in [self.dependent_variable_field_name, self.independent_variable_field_name]:
            if isinstance(column, OnetickParameter):
                continue

            if column not in src.schema:
                raise TypeError(f"Aggregation `{self.NAME}` uses column `{column}` as input, which doesn't exist")

        for out_field in ['SLOPE', 'INTERCEPT']:
            if out_field in src.schema:
                raise TypeError(f"Field `{out_field}`, which is `{self.NAME}` aggregation output column, is in schema.")

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        return {
            'SLOPE': float,
            'INTERCEPT': float,
        }


class PartitionEvenlyIntoGroups(_Aggregation, _MultiColumnAggregation):
    NAME = 'PARTITION_EVENLY_INTO_GROUPS'
    EP = otq.PartitionEvenlyIntoGroups

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['field_to_partition'] = 'FIELD_TO_PARTITION'
    FIELDS_MAPPING['weight_field'] = 'WEIGHT_FIELD'
    FIELDS_MAPPING['number_of_groups'] = 'NUMBER_OF_GROUPS'

    FIELDS_TO_SKIP = ['column_name', 'output_field_name']

    def __init__(
        self,
        field_to_partition: Union[_Column, str, OnetickParameter],
        weight_field: Union[_Column, str, OnetickParameter],
        number_of_groups: Union[int, OnetickParameter],
        *args, **kwargs,
    ):
        if isinstance(field_to_partition, _Column):
            field_to_partition = str(field_to_partition)
        if isinstance(weight_field, _Column):
            weight_field = str(weight_field)

        if number_of_groups <= 0:
            raise ValueError('Parameter `number_of_groups` value should be greater than zero')

        self.field_to_partition = field_to_partition
        self.weight_field = weight_field
        self.number_of_groups = number_of_groups

        super().__init__(_Column('TIMESTAMP'), *args, **kwargs)

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)

        for column in [self.field_to_partition, self.weight_field]:
            if isinstance(column, OnetickParameter):
                continue

            if column not in src.schema:
                raise TypeError(f"Aggregation `{self.NAME}` uses column `{column}` as input, which doesn't exist")

        for out_field in ['FIELD_TO_PARTITION', 'GROUP_ID']:
            if out_field in src.schema:
                raise TypeError(f"Field `{out_field}`, which is `{self.NAME}` aggregation output column, is in schema.")

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        return {
            'FIELD_TO_PARTITION': str,
            'GROUP_ID': int,
        }



import difflib
import os
import sysconfig
import warnings
from collections import OrderedDict

from onetick.py.otq import otq

from .temp import TmpFile, TmpDir, ONE_TICK_TMP_DIR
from ..core import db_constants
from . import types


def reload_config(db=None, config_type='LOCATOR'):
    import onetick.py as otp

    if db is not None:
        warnings.warn("Parameter 'db' is deprecated and has no meaning")

    return otp.run(
        otq.ReloadConfig(config_type=config_type).tick_type('ANY'),
        symbols='LOCAL::',
        # start and end times don't matter for this query, use some constants
        start=db_constants.DEFAULT_START_DATE,
        end=db_constants.DEFAULT_END_DATE,
    )


def _remove_quotes(s):
    return s.replace('"', "").replace("'", "")


def _read_config_params(path, _included_paths=None):
    params = OrderedDict()

    if _included_paths is None:
        _included_paths = set()

    with open(path, "r") as fin:
        for line in fin:
            line = line.strip()

            if line.startswith("#"):
                continue

            if line.startswith("INCLUDE"):
                included_path = os.path.expandvars(_remove_quotes(line.split()[1]))
                if included_path in _included_paths:
                    raise RecursionError(f"Path '{included_path}' is included more than once")
                _included_paths.add(included_path)
                included_params = _read_config_params(included_path, _included_paths)
                params.update({
                    k: v for k, v in included_params.items()
                    if k not in params
                })
                continue

            values = line.split("=")
            key, value = values[0].strip(), os.path.expandvars("".join(values[1:]).strip())

            if key and key not in params:
                params[key] = value.strip()

    return params


def _write_config_params(path, params):
    with open(path, "w") as fout:
        for key, value in params.items():
            fout.write(f"{key}={value}\n")


def _check_param_in_params(param, params, throw_exception=True):
    if param in params:
        return True
    if not throw_exception:
        return False
    closest = difflib.get_close_matches(param, params.keys())
    msg = f"Parameter '{param}' is not in the config file."
    if closest:
        msg = f"{msg} It might be you supposed '{closest[0]}'?"
    raise AttributeError(msg)


def modify_config_param(path, param, value, throw_on_missing=True):
    params = _read_config_params(path)

    _check_param_in_params(param, params, throw_on_missing)

    params[param] = f'"{value}"'

    _write_config_params(path, params)


def get_config_param(path, param, default=None):
    params = _read_config_params(path)

    res = _check_param_in_params(param, params, throw_exception=(default is None))

    if res:
        return _remove_quotes(params[param])
    return default


def is_param_in_config(path, param):
    params = _read_config_params(path)

    return _check_param_in_params(param, params, throw_exception=False)


def tmp_config(
    locator=None,
    acl=None,
    otq_path=None,
    csv_path=None,
    clean_up=types.default,
    license_path=None,
    license_dir=None,
    license_servers=None,
):
    data = []

    if otq_path is None:
        otq_path = []

    if ONE_TICK_TMP_DIR():
        otq_path.append(ONE_TICK_TMP_DIR())

    if csv_path is None:
        csv_path = []

    otq_path = list(map(os.path.normpath, otq_path))
    csv_path = list(map(os.path.normpath, csv_path))

    data.append("ONE_TICK_CONFIG.ALLOW_ENV_VARS=Yes")
    data.append("")
    config_vars = sysconfig.get_config_vars()
    data.append(f"PYTHON_VERSION='{config_vars['py_version_short']}{config_vars['abiflags']}'")

    if locator:
        data.append(f'DB_LOCATOR.DEFAULT="{_remove_quotes(str(locator))}"')

    if acl:
        data.append(f'ACCESS_CONTROL_FILE="{_remove_quotes(str(acl))}"')

    if license_path:
        data.append(f'ONE_TICK_LICENSE_FILE="{_remove_quotes(str(license_path))}"')

    if license_dir:
        data.append(f'LICENSE_REPOSITORY_DIR="{_remove_quotes(str(license_dir))}"')

    if license_servers:
        data.append(f'LICENSE_TICK_SERVERS="{_remove_quotes(str(license_servers))}"')

    if otq_path:
        data.append(f'OTQ_FILE_PATH="{",".join(otq_path)}"')

    if csv_path:
        data.append(f'CSV_FILE_PATH="{",".join(csv_path)}"')

    if os.getenv('OTP_WEBAPI_TEST_MODE'):
        tmp_dir = TmpDir()
        data.append(f'TICK_SERVER_CSV_CACHE_DIR="{tmp_dir}"')
        data.append(f'TICK_SERVER_DATA_CACHE_DIR="{tmp_dir}"')
        data.append(f'TICK_SERVER_OTQ_CACHE_DIR="{tmp_dir}"')

    data.append("")
    data.append("ALLOW_REMOTE_CONTROL=Yes")
    data.append("ALLOW_NO_CERTIFICATE=true")
    # to be able to run otp.ODBC
    data.append("LOAD_ODBC_UDF=true")

    # # let's not strip quotes around params by default, this is the old behaviour
    # new default behaviour breaks backward-compatibility (BEXRTS-1340, BEXRTS-1342)
    # new behaviour can be achieved by setting STRIP_QUOTES_WHEN_ASSIGNING_SYMBOL_AND_OTQ_PARAMS param in JWQ EP
    data.append("COMPATIBILITY.JOIN_WITH_QUERY.STRIP_QUOTES_AROUND_SYMBOL_AND_OTQ_PARAMS=FALSE")

    if os.getenv('OTP_WEBAPI_TEST_MODE'):
        tmp_file = TmpFile(name="onetick.cfg", clean_up=clean_up, force=True)
    else:
        tmp_file = TmpFile(suffix=".cfg", clean_up=clean_up)

    with open(tmp_file, "w") as fout:
        fout.write("\n".join(data))

    return tmp_file
